###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from proxies.mytorch.nn.layers_proxy import LinearProxy

class Linear:
    def __init__(self, in_features, out_features):
        self.in_features = in_features
        self.out_features = out_features
        self.proxy = LinearProxy(in_features, out_features)


from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class ModelIDandTensorID(_message.Message):
    __slots__ = ("model_uuid", "tensor_uuid")
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    tensor_uuid: str
    def __init__(self, model_uuid: _Optional[str] = ..., tensor_uuid: _Optional[str] = ...) -> None: ...

class SaveRequest(_message.Message):
    __slots__ = ("model_uuid", "path")
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    path: str
    def __init__(self, model_uuid: _Optional[str] = ..., path: _Optional[str] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.mytorch_proxy import MyTorchProxy
from proxies.no_op import NoOpContextManager
from torch.Tensor import Tensor
from torch.device import device
import numpy as np
from typing import Dict, List
import json

def randn(*size: int) -> Tensor:
    size = list(size)
    return MyTorchProxy().randn(size)

def rand(*size: int) -> Tensor:
    size = list(size)
    return MyTorchProxy().rand(size)

def zeros(*size: int) -> Tensor:
    size = list(size)
    return MyTorchProxy().zeros(size)

def ones(*size: int) -> Tensor:
    size = list(size)
    return MyTorchProxy().ones(size)

def from_numpy(ndarray: np.ndarray) -> Tensor:
    # Implementation to request a tensor from the server
    return MyTorchProxy().from_numpy(ndarray)

def no_grad():
    return NoOpContextManager()

def max(tensor: Tensor, dim: int) -> tuple[Tensor, Tensor]:
    return MyTorchProxy().max(tensor, dim)

def repeat_interleave(tensor: Tensor, repeats: int, dim: int) -> Tensor:
    return MyTorchProxy().repeat_interleave(tensor, repeats, dim)

def unsqueeze(tensor: Tensor, dim: int) -> Tensor:
    return MyTorchProxy().unsqueeze(tensor, dim)

def arange(start, end, step) -> Tensor:
    # make sure stare, end, and step are floats
    start = float(start)
    end = float(end)
    step = float(step)
    return MyTorchProxy().arange(start, end, step)

def meshgrid(*tensors: Tensor, indexing: str = 'xy') -> tuple:
    tensor_uuids = [tensor.uuid for tensor in tensors]
    return MyTorchProxy().meshgrid(tensor_uuids, indexing)

def reshape(tensor: Tensor, shape: tuple) -> Tensor:
    return MyTorchProxy().reshape(tensor, shape)

def cat(tensors, dim: int) -> Tensor:
    return MyTorchProxy().cat(tensors, dim)

def argmax(tensor: Tensor, dim: int = None, keepdim: bool = False) -> Tensor:
    return MyTorchProxy().argmax(tensor, dim, keepdim)

def matmul(tensor1: Tensor, tensor2: Tensor) -> Tensor:
    return MyTorchProxy().matmul(tensor1, tensor2)

def load(file_path: str) -> Dict[str, Tensor]:
    return MyTorchProxy().load(file_path)

def save(state_dict: Dict[str, Tensor], file_path: str):
    return MyTorchProxy().save(state_dict, file_path)

def allclose(input: Tensor, other: Tensor, rtol: float = 1e-05, atol: float = 1e-08, equal_nan: bool = False) -> bool:
    return MyTorchProxy().allclose(input, other, rtol, atol, equal_nan)

def device(dev: str) -> device:
    return MyTorchProxy().device(dev)

def tanh(tensor: Tensor) -> Tensor:
    return MyTorchProxy().tanh(tensor)

def sigmoid(tensor: Tensor) -> Tensor:
    return MyTorchProxy().sigmoid(tensor)

def stack(tensors : List, dim: int = 0) -> Tensor:
    tensor_uuids = [tensor.uuid for tensor in tensors]
    return MyTorchProxy().stack( tensor_uuids, dim)

def mean(input, dim=None, keepdim=False, dtype=None, out=None) -> Tensor:
    # If the input tensor is empty, torch.mean() returns nan. 
    # This behavior is consistent with NumPy
    return MyTorchProxy().mean(input, dim, keepdim, dtype, out)

def randn_like(input: Tensor) -> Tensor:
    return randn(*input.size())

def equal(tensor1: Tensor, tensor2: Tensor) -> bool:
    return MyTorchProxy().equal(tensor1, tensor2)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/nn/module.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2
from gRPC_impl.mytorch.nn import nn_msg_types_pb2 as mytorch_dot_nn_dot_nn__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17mytorch/nn/module.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\x1a\x1dmytorch/nn/nn_msg_types.proto\"5\n\x15ModuleToDeviceRequest\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0e\n\x06\x64\x65vice\x18\x02 \x01(\t\"$\n\nParameters\x12\x16\n\x0egenerator_uuid\x18\x01 \x01(\t\"\x1c\n\x0chalf_request\x12\x0c\n\x04uuid\x18\x01 \x01(\t\"-\n\rhalf_response\x12\x1c\n\x05\x65mpty\x18\x01 \x01(\x0b\x32\r.shared.Empty2\xe5\x03\n\rModuleService\x12:\n\x07\x46orward\x12\x1b.mytorch.ForwardPassRequest\x1a\x12.shared.GrpcTensor\x12\x33\n\x14\x43reateModuleOnServer\x12\r.shared.Empty\x1a\x0c.shared.UUID\x12\x33\n\x0cGetStateDict\x12\x0c.shared.UUID\x1a\x13.shared.NamedTensor0\x01\x12<\n\rLoadStateDict\x12\x13.shared.NamedTensor\x1a\x14.shared.BooleanValue(\x01\x12#\n\x04\x45val\x12\x0c.shared.UUID\x1a\r.shared.Empty\x12\x39\n\x08ToDevice\x12\x1e.mytorch.ModuleToDeviceRequest\x1a\r.shared.Empty\x12\x32\n\rGetParameters\x12\x0c.shared.UUID\x1a\x13.mytorch.Parameters\x12\x35\n\x04half\x12\x15.mytorch.half_request\x1a\x16.mytorch.half_response\x12%\n\x06\x64\x65lete\x12\x0c.shared.UUID\x1a\r.shared.Emptyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.nn.module_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MODULETODEVICEREQUEST']._serialized_start=91
  _globals['_MODULETODEVICEREQUEST']._serialized_end=144
  _globals['_PARAMETERS']._serialized_start=146
  _globals['_PARAMETERS']._serialized_end=182
  _globals['_HALF_REQUEST']._serialized_start=184
  _globals['_HALF_REQUEST']._serialized_end=212
  _globals['_HALF_RESPONSE']._serialized_start=214
  _globals['_HALF_RESPONSE']._serialized_end=259
  _globals['_MODULESERVICE']._serialized_start=262
  _globals['_MODULESERVICE']._serialized_end=747
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This class represents a torch.device object. It is used to specify the device on which a tensor is stored.

A torch.device can be constructed via a string or via a string and device ordinal

Via a string:
    torch.device('cuda:0')
    torch.device('cpu')

Via a string and device ordinal:

    torch.device('cuda', 0)
"""

class device:
    def __init__(self, device_str: str, device_ordinal: int = 0):
        self.device_str = device_str
        self.device_ordinal = device_ordinal

    # overwrite the __str__ method to return the device string
    def __str__(self):
        return self.device_str



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytorch.scaffolding import data_mgmt_pb2 as mytorch_dot_scaffolding_dot_data__mgmt__pb2
from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2

GRPC_GENERATED_VERSION = '1.65.0'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytorch/scaffolding/data_mgmt_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class DataMgmtServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.UploadFolder = channel.stream_unary(
                '/mytorch.DataMgmtService/UploadFolder',
                request_serializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.FileUploadRequest.SerializeToString,
                response_deserializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.FolderUploadResponse.FromString,
                _registered_method=True)
        self.FolderExistsOnServer = channel.unary_unary(
                '/mytorch.DataMgmtService/FolderExistsOnServer',
                request_serializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.RelativePath.SerializeToString,
                response_deserializer=shared__msg__types__pb2.BooleanValue.FromString,
                _registered_method=True)
        self.SetServerDataDir = channel.unary_unary(
                '/mytorch.DataMgmtService/SetServerDataDir',
                request_serializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.DataDir.SerializeToString,
                response_deserializer=shared__msg__types__pb2.Empty.FromString,
                _registered_method=True)


class DataMgmtServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def UploadFolder(self, request_iterator, context):
        """The client will create a single streaming RPC call for the session and send
        multiple FileUploadRequest messages—one for each file in the directory
        (including nested directories if needed).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FolderExistsOnServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetServerDataDir(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DataMgmtServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'UploadFolder': grpc.stream_unary_rpc_method_handler(
                    servicer.UploadFolder,
                    request_deserializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.FileUploadRequest.FromString,
                    response_serializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.FolderUploadResponse.SerializeToString,
            ),
            'FolderExistsOnServer': grpc.unary_unary_rpc_method_handler(
                    servicer.FolderExistsOnServer,
                    request_deserializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.RelativePath.FromString,
                    response_serializer=shared__msg__types__pb2.BooleanValue.SerializeToString,
            ),
            'SetServerDataDir': grpc.unary_unary_rpc_method_handler(
                    servicer.SetServerDataDir,
                    request_deserializer=mytorch_dot_scaffolding_dot_data__mgmt__pb2.DataDir.FromString,
                    response_serializer=shared__msg__types__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytorch.DataMgmtService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytorch.DataMgmtService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class DataMgmtService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def UploadFolder(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/mytorch.DataMgmtService/UploadFolder',
            mytorch_dot_scaffolding_dot_data__mgmt__pb2.FileUploadRequest.SerializeToString,
            mytorch_dot_scaffolding_dot_data__mgmt__pb2.FolderUploadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FolderExistsOnServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.DataMgmtService/FolderExistsOnServer',
            mytorch_dot_scaffolding_dot_data__mgmt__pb2.RelativePath.SerializeToString,
            shared__msg__types__pb2.BooleanValue.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetServerDataDir(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.DataMgmtService/SetServerDataDir',
            mytorch_dot_scaffolding_dot_data__mgmt__pb2.DataDir.SerializeToString,
            shared__msg__types__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from setuptools import setup, find_packages

setup(
    name="mytorch-ai",
    version="0.4.2",
    description="Drop-in replacement for PyTorch that does distributed training and inference on a remote server",
    include_package_data=True,
    license="Custom License",
    long_description=open('README.md').read(),  # Note: path changed
    long_description_content_type='text/markdown',
    author="mytorch.net",
    author_email='pypi@mytorch.net',
    url="https://mytorch.net",
    classifiers=[
       "Development Status :: 4 - Beta",
       "Intended Audience :: Developers",
       "Intended Audience :: Science/Research",
       "Programming Language :: Python :: 3.10",
       "Topic :: Scientific/Engineering :: Artificial Intelligence",
       "License :: Other/Proprietary License",
    ],
    python_requires=">=3.10",
    install_requires=[
        "setuptools",
        "grpcio>=1.71.0",
        "grpcio-tools>=1.71.0",
        "numpy",
        "requests",
        "pillow",
        "tqdm",
        "matplotlib",
        "scikit-learn",
        "huggingface-hub"
    ],
    py_modules=["MyTorchClient"],
    packages=find_packages(
        include=[
            "torch*",
            "transformers*",
            "torchvision*",
            "proxies*", 
            "gRPC_impl*",
            "utils*",
            "connection_utils*",
        ],
        exclude=[
            "venv*",
            "venv.*",
            "*.venv*"
        ]
    ),
    package_data={
        '': ['LICENSE','README.md'],
        'torch': ['*.py'],
        'transformers': ['*.py'],
        'torchvision': ['*.py'],
        'proxies': ['*.py'],
        'gRPC_impl': ['*.py'],
        'utils': ['*.py'],
        'connection_utils': ['*.py'],
    },
) 



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch.nn import loss_functions_pb2_grpc, loss_functions_pb2
from proxies.mytorch.scaffolding.scaffolding_proxy import ScaffoldingProxy
from torch.Tensor import Tensor
from utils.logger import Logger
import numpy as np

class LossFunctionProxy:
    def __init__(self, loss_function_type: loss_functions_pb2.LossFunctionType):
        self.channel = ServerConnection.get_active_connection()
        self.stub = loss_functions_pb2_grpc.LossFunctionServiceStub(self.channel)
        self.logger = Logger.get_logger()
        # Create a LossFunctionRequest object
        request = loss_functions_pb2.LossFunctionRequest()
        request.loss_function_type = loss_function_type
        self.loss_function_uuid = self.stub.CreateLossFunction(request).uuid
        self.loss_function_type = loss_function_type

    @classmethod
    def create_CrossEntropyLoss(cls):
        return cls(loss_functions_pb2.CrossEntropy)

    @wrap_with_error_handler
    def run_loss(self, input_data, target) -> Tensor:
        # for some reason, right now the train_test_split function
        # is returning a list of numpy arrays, so convert it here
        # TODO: fix this in the train_test_split function
        if isinstance(input_data, np.ndarray):
            input_uuid = MyTorchProxy().from_numpy(input_data).uuid
        else:
            input_uuid = input_data.uuid

        if isinstance(target, np.ndarray):
            target_uuid = MyTorchProxy().from_numpy(target).uuid
        else:
            target_uuid = target.uuid

        request = loss_functions_pb2.RunLossFunctionRequest()
        request.loss_function_uuid = self.loss_function_uuid
        request.input_uuid = input_uuid
        request.target_uuid = target_uuid

        response = self.stub.RunLossFunction(request)
        return Tensor(response.uuid, response.shape, response.dtype)




from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class encodePlus_request(_message.Message):
    __slots__ = ("tokenizer_uuid", "text", "kwargs")
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    TEXT_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    text: str
    kwargs: _containers.RepeatedCompositeFieldContainer[_shared_msg_types_pb2.keyValuePair]
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., text: _Optional[str] = ..., kwargs: _Optional[_Iterable[_Union[_shared_msg_types_pb2.keyValuePair, _Mapping]]] = ...) -> None: ...

class tokenizer_key_uuid_pair(_message.Message):
    __slots__ = ("key", "tensor_uuid")
    KEY_FIELD_NUMBER: _ClassVar[int]
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    key: str
    tensor_uuid: str
    def __init__(self, key: _Optional[str] = ..., tensor_uuid: _Optional[str] = ...) -> None: ...

class encodePlus_response(_message.Message):
    __slots__ = ("encoding_uuid", "data")
    ENCODING_UUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    encoding_uuid: str
    data: _containers.RepeatedCompositeFieldContainer[tokenizer_key_uuid_pair]
    def __init__(self, encoding_uuid: _Optional[str] = ..., data: _Optional[_Iterable[_Union[tokenizer_key_uuid_pair, _Mapping]]] = ...) -> None: ...

class setPadToken_request(_message.Message):
    __slots__ = ("tokenizer_uuid", "pad_token")
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    PAD_TOKEN_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    pad_token: str
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., pad_token: _Optional[str] = ...) -> None: ...

class setPadToken_response(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class decode_request(_message.Message):
    __slots__ = ("tokenizer_uuid", "input_tensor_uuid", "kwargs")
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    INPUT_TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    input_tensor_uuid: str
    kwargs: _containers.RepeatedCompositeFieldContainer[_shared_msg_types_pb2.keyValuePair]
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., input_tensor_uuid: _Optional[str] = ..., kwargs: _Optional[_Iterable[_Union[_shared_msg_types_pb2.keyValuePair, _Mapping]]] = ...) -> None: ...

class decode_response(_message.Message):
    __slots__ = ("text",)
    TEXT_FIELD_NUMBER: _ClassVar[int]
    text: str
    def __init__(self, text: _Optional[str] = ...) -> None: ...

class tokenizerFromPretrained_request(_message.Message):
    __slots__ = ("pretrained_model_name_or_path", "inputs", "kwargs")
    class KwargsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    PRETRAINED_MODEL_NAME_OR_PATH_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    pretrained_model_name_or_path: str
    inputs: _containers.RepeatedScalarFieldContainer[str]
    kwargs: _containers.ScalarMap[str, str]
    def __init__(self, pretrained_model_name_or_path: _Optional[str] = ..., inputs: _Optional[_Iterable[str]] = ..., kwargs: _Optional[_Mapping[str, str]] = ...) -> None: ...

class tokenizerFromPretrained_response(_message.Message):
    __slots__ = ("tokenizer_uuid", "special_tokens")
    class SpecialTokensEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    SPECIAL_TOKENS_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    special_tokens: _containers.ScalarMap[str, str]
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., special_tokens: _Optional[_Mapping[str, str]] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This proxy provides a proxy for the Module gRPC service.
It allows the client to call specified torch.nn.Module operations on the server via gRPC.
"""

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl import shared_msg_types_pb2
from gRPC_impl.mytorch import tensor_pb2_grpc, tensor_pb2
from utils.data_transform_utils import deserialize_tensor
from utils.logger import Logger
import numpy as np
import json


from proxies.base_proxy import BaseProxy
class TensorProxy (BaseProxy):

    # Note that is is NOT used to create a tensor on the server side.
    # There are separate static methods for this.
    @wrap_with_error_handler
    def __init__(self, uuid: str):
        self.channel = ServerConnection.get_active_connection()
        self.stub = tensor_pb2_grpc.TensorServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.uuid = uuid

    @wrap_with_error_handler
    def get_data(self) -> np.ndarray:
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        response: shared_msg_types_pb2.SerializedTensorData = self.stub.get_data(request)
        deserialized_tensor = deserialize_tensor(response.data, response.shape, response.dtype)
        return deserialized_tensor

    @wrap_with_error_handler
    def backward(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        return self.stub.backward(request)

    @wrap_with_error_handler
    def item(self) -> float:
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        return self.stub.item(request).value

    @wrap_with_error_handler
    def float(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        grpc_tensor = self.stub.float(request)
        # can't return a mytorch.Tensor here because it would create a circular import
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    @wrap_with_error_handler
    def long(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        grpc_tensor = self.stub.long(request)
        # can't return a mytorch.Tensor here because it would create a circular import
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    @wrap_with_error_handler
    def toCuda(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        grpc_tensor = self.stub.to_cuda(request)
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    @wrap_with_error_handler
    def toCpu(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        grpc_tensor = self.stub.to_cpu(request)
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    @wrap_with_error_handler
    def reshape(self, shape: tuple):
        request = shared_msg_types_pb2.ReshapeRequest(tensor_uuid=self.uuid, shape=shape)
        grpc_tensor = self.stub.reshape(request)
        # can't return a mytorch.Tensor here because it would create a circular import
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    @wrap_with_error_handler
    def get_shape(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        grpc_tensor = self.stub.get_shape(request)
        # convert the shape from a repeated field to a tuple
        shape = tuple(grpc_tensor.shape)
        return shape

    @wrap_with_error_handler
    def get_dtype(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        grpc_tensor = self.stub.get_dtype(request)
        return grpc_tensor.dtype

    @wrap_with_error_handler
    def equal(self, other_uuid: str):
        if not isinstance(other_uuid, str):
            raise ValueError("other_uuid must be a string")
        if self.uuid == other_uuid:
            return True
        request = shared_msg_types_pb2.TwoTensorIDs(tensor1_uuid=self.uuid, tensor2_uuid=other_uuid)
        grpc_tensor = self.stub.equal(request)
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    @wrap_with_error_handler
    def add(self, other_uuid: str):
        request = shared_msg_types_pb2.TwoTensorIDs(tensor1_uuid=self.uuid, tensor2_uuid=other_uuid)
        grpc_tensor = self.stub.add(request)
        # can't return a mytorch.Tensor here because it would create a circular import
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype


    @wrap_with_error_handler
    def sum(self, operand):
        return self.generic_call("torch.Tensor", "sum", self.uuid, operand, call_type="method")

    @wrap_with_error_handler
    def sub(self, operand: int):
        return self.generic_call("torch.Tensor", "sub", self.uuid, operand, call_type="method")

    @wrap_with_error_handler
    def mul(self, operand: int):
        return self.generic_call("torch.Tensor", "mul", self.uuid, operand, call_type="method")

    @wrap_with_error_handler
    def pow(self, operand: int):
        return self.generic_call("torch.Tensor", "pow", self.uuid, operand, call_type="method")

    @wrap_with_error_handler
    def view(self, batch_size: int, input_size: int):
        return self.generic_call("torch.Tensor", "view", self.uuid, batch_size, input_size, call_type="method")
   
    @wrap_with_error_handler
    def t(self):
        return self.generic_call("torch.Tensor", "t", self.uuid, call_type="method")
   
    @wrap_with_error_handler
    def chunk(self, chunks: int, dim: int):
        """
        Calls a 'chunk' method on the server that splits the tensor
        into the specified number of chunks along the given dimension.
        Returns enough information to reconstruct each resulting chunk
        as a mytorch.Tensor on the client side.
        """
        response = self.generic_call("torch.Tensor", "chunk", self.uuid, chunks, dim, call_type="method")
        
        chunked_tensors = []
        for t_def in response:
            chunk_uuid = t_def[0] #"uuid"]
            chunk_shape = t_def[1] # "shape"]
            chunk_dtype = t_def[2] #"dtype"]
            chunked_tensors.append( (chunk_uuid, chunk_shape, chunk_dtype))
        
        return chunked_tensors


    @wrap_with_error_handler
    def slice(self, slicing_specs):
        slice_request = tensor_pb2.SliceRequest()
        slice_request.tensor_uuid = self.uuid

        # If slicing_specs is not a tuple, make it one for consistent handling
        if not isinstance(slicing_specs, tuple):
            slicing_specs = (slicing_specs,)
        
        # Process each dimension's slice specification
        for slice_spec in slicing_specs:
            tensor_slice = tensor_pb2.TensorSlice()
            
            # Handle integer index
            if isinstance(slice_spec, int):
                # For integer index i, it's equivalent to slice(i, i+1, 1)
                tensor_slice.start = slice_spec
                tensor_slice.stop = slice_spec + 1
                tensor_slice.step = 1
            
            # Handle slice object
            elif isinstance(slice_spec, slice):
                tensor_slice.start = slice_spec.start if slice_spec.start is not None else -1
                tensor_slice.stop = slice_spec.stop if slice_spec.stop is not None else -1
                tensor_slice.step = slice_spec.step if slice_spec.step is not None else 1
            
            else:
                raise TypeError(f"Unsupported slice specification type: {type(slice_spec)}")
            
            slice_request.slices.append(tensor_slice)

        grpc_tensor = self.stub.slice(slice_request)
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    # e.g. tensor[0]
    @wrap_with_error_handler
    def index(self, idx: int):
        request = shared_msg_types_pb2.TensorIDAndDim(tensor_uuid=self.uuid, dim=idx)
        grpc_tensor = self.stub.index(request)
        # can't return a mytorch.Tensor here because it would create a circular import
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    # Compare OLD and NEW versions of (un)squeeze: 
    @wrap_with_error_handler
    def squeeze(self, dim: int):
        return self.generic_call("torch.Tensor", "squeeze", self.uuid, dim, call_type="method")
   
    @wrap_with_error_handler
    def unsqueeze(self, dim: int):
        request = shared_msg_types_pb2.TensorIDAndDim(tensor_uuid=self.uuid, dim=dim)
        grpc_tensor = self.stub.unsqueeze(request)
        # can't return a mytorch.Tensor here because it would create a circular import
        return grpc_tensor.uuid, grpc_tensor.shape, grpc_tensor.dtype

    ### KEY METHOD: When objects go out of scope, this is called to clean up server memory.
    # See also ModuleProxy.delete
    @wrap_with_error_handler
    def delete(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        self.stub.delete(request)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from torch.nn.Module import Module
from proxies.mytorch.nn.sequential_proxy import SequentialProxy
from proxies.mytorch.nn.lstm_proxy import LSTMProxy
from proxies.mytorch.nn.layers_proxy import LinearProxy, ReLUProxy, FlattenProxy
from proxies.mytorch.nn.layers_proxy import LinearProxy, ReLUProxy, FlattenProxy, SigmoidProxy
import torch.nn as nn

class Sequential(Module):
    def __init__(self, *args):
        super().__init__(uuid="0000")
        layer_proxies = [self._convert_to_proxy(layer) for layer in args]
        self.proxy = SequentialProxy(*layer_proxies)
        self.uuid = self.proxy.create_sequential_on_server() 
        self.set_uuid(self.uuid)

    def forward(self, input_data):
        return super().forward(input_data)

    def _convert_to_proxy(self, layer):
        if isinstance(layer, nn.Linear):
            return LinearProxy(layer.in_features, layer.out_features)
        elif isinstance(layer, nn.ReLU):
            return ReLUProxy()
        elif isinstance(layer, nn.Flatten):
            return FlattenProxy()

        elif isinstance(layer, nn.LSTM):
            return LSTMProxy()
        elif isinstance(layer, nn.Sigmoid):
            return SigmoidProxy()
        else:
            raise NotImplementedError(f"Layer type {type(layer)} is not supported.")




###############################################################################
# Copyright (c) 2024-2025 MyTorch Systems Inc. All rights reserved.
###############################################################################
from torch.utils.data.Dataset import Dataset
from torch.Tensor import Tensor
from typing import Tuple, Sequence
from proxies.mytorch.utils.data.tensordataset_proxy import TensorDatasetProxy


class TensorDataset(Dataset):
    def __init__(self, dataset: Dataset) -> None:
        super().__init__()
        self.dataset = dataset
        self.proxy = TensorDatasetProxy()
        self.uuid, self.dataset_length = self.createTensorDatasetOnServer()

    def createTensorDatasetOnServer(self):
        uuid, dataset_length = self.proxy.createTensorDatasetOnServer(self.dataset)
        return uuid, dataset_length

    def __len__(self):
        return self.dataset_length



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch_neuronx.mytorch_neuronx_proxy import CqtorchNeuronxProxy
from torch.Tensor import Tensor
from torch.nn.Module import Module

'''
FROM: help(torch_neuronx.trace)
Help on function trace in module torch_neuronx.xla_impl.trace:

trace(func: Union[Callable, torch.nn.modules.module.Module], 
        example_inputs: Any, *_,  <-- must be same dimensions as the input(s) you will pass to the model
        input_output_aliases: Dict = {}, 
        compiler_workdir: Union[str, pathlib.Path, NoneType] = None, 
        compiler_args: Union[List[str], str, NoneType] = None, 
        partitioner_config: Optional[torch_neuronx.xla_impl.partitioner.PartitionerConfig] = None, 
        inline_weights_to_neff=True, **kwargs) 
        -> torch.jit._script.ScriptModule
    Trace a torch module/function to produce a compiled Neuron ScriptModule.
    
    This uses torch-xla to extract the computation graph. The input `func` and
    `example_inputs` must be able to be moved to the XLA device.
    
    The resulting module can used with `torch.jit.save` and `torch.jit.load`
    
    NOTE: Please use keyword arguments for all arguments after `example_inputs`.
    Ex: `torch_neuronx.trace(func,example_inputs,compiler_workdir="some_dir") #valid`
        `torch_neuronx.trace(func,example_inputs,"some_dir") #invalid`
    
    Args:
        func: A module or function which defines a torch model or computation.
        example_inputs: An example set of inputs which will be passed to the
            `func` during tracing.
        input_output_aliases: alias mapping between the inputs and outputs
        compiler_workdir: The directory to save any compiler outputs to.
        compiler_args: Additional compiler arguments.
        partitioner_config: A PartitionerConfig object, which can be optionally
        supplied if there are unsupported ops in the model that need to be
        partitioned out to CPU.
        inline_weights_to_neff: A boolean indicating whether the weights should be
        inlined to the neff. The default is `True`.
    
    Returns:
        A Module where the HLO computation is a fused neuron::foward operation.
'''
def trace(model:Module, example_inputs: tuple[Tensor, ...]):
    model_uuid =  CqtorchNeuronxProxy().trace(model, example_inputs)
    return Module(uuid=model_uuid)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This singleton class gets the gRPC connection to the server.
"""
import grpc
from utils.logger import Logger
import time
from utils.MyTorchClient import MyTorchClient
import signal
import os
import sys
import atexit
import threading

###########################################################################
#### ServerConnection; only has one connection to the server
#### All methods are class methods.
#### _connection_is_active is thread safe; if True, all subsequent calls are ignored
###########################################################################

class ServerConnection:
    _lock = threading.Lock() # lock to ensure that only one thread can access the connection at a time
    _channel = None
    _server = None
    _port = 50051 # default
    _connection_is_active = False # set to True when disconnect is called so all subsequent calls are ignored

    @classmethod
    def get_server_ip(cls):
        return cls._server, cls._port

    @classmethod
    def close_connection(cls):
        with cls._lock:
            cls._connection_is_active = True

    @classmethod
    def open_connection(cls):
        with cls._lock:
            cls._connection_is_active = False

    @classmethod
    def is_connection_active(cls):
        with cls._lock:
            return cls._connection_is_active

    @classmethod
    def check_mytorch_token(cls):
        """
        Check for MyTorch token in environment variables or config file.
        Raises RuntimeError if token is not found.
        """
        # First check environment variable
        token = os.getenv('MYTORCH_TOKEN')
        if token:
            return token

        # Check ~/.mytorch file
        config_path = os.path.expanduser('~/.mytorch')
        if os.path.exists(config_path):
            with open(config_path, 'r') as f:
                for line in f:
                    if 'token' in line and '=' in line:
                        # Split on '=' and strip whitespace from both parts
                        parts = line.split('=', 1)
                        if parts[0].strip() == 'token':
                            token = parts[1].strip()
                            # Set the environment variable for future use
                            os.environ['MYTORCH_TOKEN'] = token
                            return token

        print("*** ERROR -- MyTorch token not found.\n"
              "*** Please add 'token=<your_token>' to ~/.mytorch file or set the MYTORCH_TOKEN environment variable", file=sys.stderr)
        sys.exit(1)

    @classmethod
    def get_active_connection(cls):
        """
        If there is no connection, connect to the server and register client
        If connection is active, return the existing connection
        """
        if cls._channel is None:
            # Check for token before attempting connection
            cls.check_mytorch_token()

            # if the server is specified in the environment variable, use that, otherwise use the proxy
            if os.getenv('MYTORCH_SERVER_IP') is not None:
                cls._server, cls._port = MyTorchClient.instance().select_server_by_env_var()
            else:
                response = MyTorchClient.instance().select_server_by_proxy()
                if response is not None:
                    cls._server, cls._port = response
                else:
                    Logger.get_logger().error("Invalid access token or failed to connect to mytorch server")
                    sys.exit(1)

            # Create a channel with the server
            channel = grpc.insecure_channel(cls._server + ":" + str(cls._port))
            # Define the metadata to be added
            client_id = MyTorchClient.instance().get_client_id()
            metadata = [('client_id', client_id)]

            register_cleanup_handlers()

            """
            ... currently don't need this, but this is how you would add more metadata
            to the connection.
            
            client_info = MyTorchClient.instance().get_client_info()
            metadata.extend([
                ('hostname', client_info['hostname']),
                ('local_ip', client_info['local_ip']),
                ('os_name', client_info['os_name']),
                ('os_version', client_info['os_version']),
                ('os_release', client_info['os_release']),
                ('os_details', client_info['os_details']),
                ('mytorch_version', client_info['mytorch_version']),
                ('machine', client_info['machine']),
                ('processor', client_info['processor']),
                ('token', client_info['token']),
            """

            metadata_interceptor = MetadataInterceptor(metadata)
            logging_interceptor = LoggingInterceptor()

            # Add the interceptors to the channel
            try:
                cls._channel = grpc.intercept_channel(channel, logging_interceptor, metadata_interceptor)
                cls.open_connection()
                Logger.get_logger().info(f"Connecting to server {cls._server}:{cls._port}")
            except Exception as e:
                Logger.get_logger().error(f"Error connecting to server {cls._server}:{cls._port}")
                Logger.get_logger().error(f"Error details: {e}")
                sys.exit(1)

            # if we get here, we are connected to the server; now we need to register the client
            client_info = MyTorchClient.instance().get_client_info()
            from proxies.mytorch.scaffolding.scaffolding_proxy import ScaffoldingProxy
            response = ScaffoldingProxy().register_client_with_server(client_info)
            if response is not None and str(response) != "":
                Logger.get_logger().error(f"Error registering client with server: {response}")
                sys.exit(1)
        return cls._channel

###########################################################################
#### Interceptor to add metadata (e.g. Client UUID) to each outgoing request
###########################################################################
class MetadataInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor
):
    def __init__(self, metadata):
        self._metadata = metadata

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details = self._augment_metadata(client_call_details)
        return continuation(new_details, request)

    def intercept_unary_stream(self, continuation, client_call_details, request):
        new_details = self._augment_metadata(client_call_details)
        return continuation(new_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        new_details = self._augment_metadata(client_call_details)
        return continuation(new_details, request_iterator)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        new_details = self._augment_metadata(client_call_details)
        return continuation(new_details, request_iterator)

    def _augment_metadata(self, client_call_details):
        new_metadata = []
        if client_call_details.metadata is not None:
            new_metadata = list(client_call_details.metadata)
        new_metadata.extend(self._metadata)

        # Use the named tuple to create new client call details
        return _ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            credentials=client_call_details.credentials,
            metadata=new_metadata,
            wait_for_ready=client_call_details.wait_for_ready,
            compression=client_call_details.compression
        )

# Define a named tuple to represent the client call details
from collections import namedtuple
_ClientCallDetails = namedtuple(
    '_ClientCallDetails',
    ['method', 'timeout', 'credentials', 'metadata', 'wait_for_ready', 'compression']
)

###########################################################################
#### Interceptor to add client-side logging of gRPC calls
###########################################################################
class LoggingInterceptor(grpc.UnaryUnaryClientInterceptor,
                         grpc.UnaryStreamClientInterceptor,
                         grpc.StreamUnaryClientInterceptor,
                         grpc.StreamStreamClientInterceptor):
    def __init__(self):
        self.logger = Logger.get_logger()

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_unary_stream(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_stream_call(continuation, client_call_details, request_iterator)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        return self._intercept_stream_call(continuation, client_call_details, request_iterator)

    def _intercept_call(self, continuation, client_call_details, request):
        # this will be intercepted by the @wrap_with_error_handler decorator and ignored
        if ServerConnection.is_connection_active():
            raise IgnoredGrpcCallException("Connection already closed; ignoring gRPC call.")
        method_name = client_call_details.method
        self.logger.debug(f"~~~~~~~~~ Outgoing request ~~~~~~~~~")
        self.logger.debug(f"Calling gRPC method: {method_name}")

        start_time = time.perf_counter()
        response = continuation(client_call_details, request)
        self._log_duration(method_name, start_time)
        self._log_response(method_name, response)
        return response

    def _intercept_stream_call(self, continuation, client_call_details, request_iterator):
        # this will be intercepted by the @wrap_with_error_handler decorator and ignored
        if ServerConnection.is_connection_active():
            raise IgnoredGrpcCallException("Connection already closed; ignoring gRPC call.")

        method_name = client_call_details.method
        self.logger.debug(f"~~~~~~~~~ Outgoing stream request ~~~~~~~~~")
        self.logger.debug(f"Calling gRPC streaming method: {method_name}")

        start_time = time.perf_counter()

        # Wrap the request iterator to add logging
        wrapped_iterator = self._log_stream_requests(request_iterator)

        response = continuation(client_call_details, wrapped_iterator)
        self._log_duration(method_name, start_time)
        self._log_response(method_name, response)
        return response

    def _log_duration(self, method_name, start_time):
        end_time = time.perf_counter()
        duration = end_time - start_time
        rounded_duration = round(1000 * duration, 2)  # Convert to milliseconds and round to 2 decimal places
        self.logger.debug(f"Method {method_name} took: {rounded_duration} ms")
        MyTorchClient.instance().add_method_duration(method_name, duration)
        return duration

    def _log_response(self, method_name, response):
        self.logger.debug("~~~~~~~~~ Incoming response ~~~~~~~~~")
        self.logger.debug(f"Received gRPC method: {method_name}")
        self.logger.debug("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
        return response

    def _log_stream_requests(self, request_iterator):
        for request in request_iterator:
            self.logger.debug(f"Streaming request: {request}")
            yield request

###########################################################################
#### this provides a central place to handle gRPC errors
###########################################################################
# Custom exception to indicate that the gRPC call should be ignored.
class IgnoredGrpcCallException(Exception):
    pass


def wrap_with_error_handler(method):
    from gRPC_impl.shared_msg_types_pb2 import Empty
    logger = Logger.get_logger()

    def wrapper(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except IgnoredGrpcCallException:  # currently only used for connection close
            return Empty()  # Return a default response or an appropriate response for your method
        except grpc.RpcError as e:
            #import traceback

            #stack_trace = traceback.format_exc()
            server, port = ServerConnection.get_server_ip()
            status_code = e.code()
            logger.error("*** Server error ***")
            logger.error(f"Failed while invoking client method: {method.__module__}.{method.__name__}")
            #logger.error(stack_trace)
            if status_code == grpc.StatusCode.UNAVAILABLE:
                logger.error(f"Please ensure the server is active and reachable.")
                sys.exit(1)
            elif status_code == grpc.StatusCode.INVALID_ARGUMENT:
                logger.error(f"Invalid argument provided to gRPC method {method.__name__}")
                sys.exit(1)
            else:
                logger.error(e.details())
                sys.exit(1)
        except Exception as e:
            if not isinstance(e, grpc.RpcError):  # Ensure it's not a grpc.RpcError
                logger.error(f"Error calling server: {method.__module__}.{method.__name__}")
                logger.error(f"Details: {str(e)}")
                sys.exit(1)

    return wrapper



###########################################################################
#### Handle client kill/end by telling server we are disconnecting
###########################################################################
# this is called when the client is cleanly disconnected
def disconnect():
    try:
        grpc_disconnect()
        Logger.get_logger().info("Disconnected from server.")
    except:
        pass  # Silently handle any errors during disconnect
    finally:
        ServerConnection.close_connection()

@ wrap_with_error_handler
def grpc_disconnect():
    from torch.scaffolding.server_mgmt import client_disconnect
    client_disconnect()

# This is called when the client is forced-quit by the user
def handle_signal(signal, frame):
    Logger.get_logger().info("Forced-quitting; disconnecting from server...")
    disconnect()
    sys.exit(0)

def register_cleanup_handlers():
    # Register signal handlers
    # call disconnect() when the client is forced-quit by the user
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)

    # Register atexit function
    # call disconnect() when the client is cleanly disconnected
    Logger.get_logger().info("Client process is exiting; disconnecting from server...")
    atexit.register(disconnect)




###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from utils.MyTorchClient import MyTorchClient
from proxies.mytorch.scaffolding.scaffolding_proxy import ScaffoldingProxy
from connection_utils.server_resolver import ServerResolver
import datetime

SERVER_LOCALHOST = 'LOCALHOST'
SERVER_AWS = 'AWS'
SERVER_TITAN = 'TITAN'
SERVER_FUNC2 = 'func2'
SERVER_CACHEQ_AI = 'mytorch.scaffolding.ai'
SERVER_NEURON = 'neuron'

def print_timing_data(summary_only: bool = False):
    MyTorchClient.instance().print_timing_data(summary_only)
    
def get_timing_data():
    return MyTorchClient.instance().get_timing_data()

### TO DO - for timing data: 
### Assume 1 current_run per client, many conections per server
### End_run should set current_run to None, deal with it.
### Also on the server, scaffolding_server_get_timing_statistics, delete data for run.
### Assume data only fetched once at end_run (so clean up memory).
### Also it would be good to delete old timing data if never fetched.
### Check for these assumptions and warn if not met.
### Client TimingData keeps data for many runs, for later graphing.
### The run_id name could encode more in engish string, cpu or gpu for example.
### Could also hold codes for different GPU types "AWS GPU_M100" for example.
### IF begin or end run methods not called, it's ok, this is optional timing data.
### Check / test these edge conditions. 

def begin_run(run_id=None):
    current_time = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    if run_id is None:    
        run_id = current_time
    else:
        run_id += "::" + current_time
    MyTorchProxy().scaffolding_server_initialize_timing_statistics(run_id)
    MyTorchClient.instance().begin_run(run_id)
    return run_id

def end_run(run_id=None):
    if run_id is None:
        run_id = MyTorchClient.instance().get_current_run_id()
    stats_json = MyTorchProxy().scaffolding_server_get_timing_statistics(run_id)
    MyTorchClient.instance().end_run(run_id, stats_json)

def select_server(server, port=50051) -> tuple[str, str, int]: # server_name, server_ip, server_port
    server_info = MyTorchClient.instance().select_server(server, port)
    client_info = MyTorchClient.instance().get_client_info()
    ScaffoldingProxy().initialize_client_connection(client_info)
    return server_info

def select_server_by_ip(ip, port=50051):
    server_info = MyTorchClient.instance().select_server_by_ip(ip, port)
    client_info = MyTorchClient.instance().get_client_info()
    ScaffoldingProxy().initialize_client_connection(client_info)
    return server_info

def select_server_by_env_var():
    server_info = MyTorchClient.instance().select_server_by_env_var()
    client_info = MyTorchClient.instance().get_client_info()
    ScaffoldingProxy().register_client_with_server(client_info)

def get_server_name():
    return MyTorchClient.instance().get_server_name()

def get_server():
    return MyTorchClient.instance().get_server()

def connect_via(dns_name) -> tuple[str, str, int]: # server_name, server_port, server_ip
    server_resolver = ServerResolver(dns_name)
    return server_resolver.connect_to_server()

### CACHEQ server-side methods:

def scaffolding_server_get_timing_statistics(run_id):
    stats_json = MyTorchProxy().scaffolding_server_get_timing_statistics(run_id)
    MyTorchClient.instance().set_server_timing_statistics(stats_json)
    return stats_json

def scaffolding_server_initialize_timing_statistics(run_id):
    MyTorchProxy().scaffolding_server_initialize_timing_statistics(run_id)

def set_overwrite_server_data(overwrite: bool):
    MyTorchClient.instance().set_overwrite_server_data(overwrite)

def get_gpu_info():
    return MyTorchProxy().get_gpu_info()


    


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from torchvision.models.weight_enums import WeightsEnum
from gRPC_impl.mytorchvision.models import models_pb2, models_pb2_grpc

class ModelsProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.logger = Logger.get_logger()
        self.stub = models_pb2_grpc.ModelsServiceStub(self.channel)

    @wrap_with_error_handler
    def resnet18(self, weights: WeightsEnum):
        request = models_pb2.WeightEnumString(value=weights.name)
        response = self.stub.resnet18(request)
        return response.uuid

    @wrap_with_error_handler
    def resnet50(self, weights: WeightsEnum):
        request = models_pb2.WeightEnumString(value=weights.name)
        response = self.stub.resnet50(request)
        return response.uuid

    @wrap_with_error_handler
    def resnet152(self, weights: WeightsEnum):
        request = models_pb2.WeightEnumString(value=weights.name)
        response = self.stub.resnet152(request)
        return response.uuid


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorchvision.datasets import datasets_pb2, datasets_pb2_grpc
from gRPC_impl.mytorch.scaffolding import data_mgmt_pb2_grpc


class ImageFolderProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.relative_datapath = None
        self.transform_uuid = None
        self.datasets_stub = datasets_pb2_grpc.DatasetsServiceStub(self.channel)
        self.data_mgmt_stub = data_mgmt_pb2_grpc.DataMgmtServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def createImageFolderOnServer(self, relative_datapath: str, transform_uuid: str):
        self.relative_datapath = relative_datapath
        self.transform_uuid = transform_uuid
        request = datasets_pb2.CreateImageFolderRequest(relative_datapath=self.relative_datapath, transform_uuid=self.transform_uuid)
        response: datasets_pb2.GrpcImageFolder = self.datasets_stub.CreateImageFolderOnServer(request)
        return response.uuid, response.image_uuids, response.labels, response.classes


# Abstract base class for proxy objects, to implement the generic_call approach.
import json
import numpy as np

from connection_utils.server_connection import wrap_with_error_handler
from gRPC_impl import shared_msg_types_pb2
from utils.logger import Logger
from gRPC_impl.mytorch import mytorch_pb2_grpc
from typing import Optional

class BaseProxy():
    def __init__(self):
        self.logger = Logger.get_logger()

    # GENERIC_CALL
    # Every call to generic_call(...) includes:
    # A standardized context (e.g., "torch", "torch.nn", "torch.Tensor")
    # An explicit and correct call_type (constructor, method, forward, function)
    @wrap_with_error_handler
    def generic_call(self, context: str, method: str, *args, call_type: str = "function", kwargs: Optional[dict] = None):
        """Handles a generic JSON request and returns a dynamic response."""
        print(f"Sending generic command: {context}: {method} {call_type} with args {args}, kwargs {kwargs}")
        #traceback.print_stack()
  
        method_json = json.dumps({
            "context": context,
            "method": method,
            "call_type": call_type,
            "args": args,
            "kwargs": kwargs or {}  
        })
    
        # Create a gRPC request
        request = shared_msg_types_pb2.JsonRequest()
        request.json_payload = method_json

        stub = mytorch_pb2_grpc.MyTorchServiceStub(self.channel)

        response: shared_msg_types_pb2.JsonResponse = stub.generic_call(request)

        # Decode JSON response
        result = json.loads(response.json_payload)

        # result could be just a simple type like the boolean True
        if isinstance(result, dict) and "error" in result:
            raise RuntimeError(f"Server Error: {result['error']}")

        return self._deserialize_response(result)

    def _deserialize_response(self, response: dict):
        """Converts a top-level JSON dict into the final Python object."""
        return self._deserialize_anything(response)


    def _deserialize_anything(self, resp_data):
        """
        Recursively convert the JSON response structure back into a Python
        object (e.g., (uuid, shape, dtype) for Tensors, 
        or nested lists, or ...).
        """

        # If it's a list, parse each item
        if isinstance(resp_data, list):
            return [self._deserialize_anything(item) for item in resp_data]

        # If it's not a dict, just return it as-is (e.g. int, float, str, bool)
        if not isinstance(resp_data, dict):
            return resp_data

        # If it is a dict, see if there's a 'type' field
        if "type" not in resp_data:
            # It's just a normal dictionary with no "type" – recursively parse each value
            return {k: self._deserialize_anything(v) for k, v in resp_data.items()}

        # Now we have a dict that has 'type'
        resp_type = resp_data["type"]

        if resp_type == "tensor":
            return (resp_data["uuid"], resp_data["shape"], resp_data["dtype"])

        elif resp_type == "module":
            return resp_data["uuid"]

        elif resp_type == "optimizer":
            return resp_data["uuid"]
        
        elif resp_type == "dataset":
            return (resp_data["uuid"], resp_data["dataset_length"])

        elif resp_type == "dataloader":
            return (
                resp_data["uuid"],
                resp_data["dataset_length"],
                resp_data["batch_size"]
            )

        elif resp_type == "scalar":
            return resp_data["value"]

        elif resp_type == "bool":
            return bool(resp_data["value"])

        elif resp_type == "list":
            # e.g. {"type": "list","value":[...]}
            return [self._deserialize_anything(item) for item in resp_data["value"]]

        # If there's some known or unknown type, handle or fallback:
        # e.g. "unknown", "dict", or others
        elif resp_type == "unknown":
            return resp_data["value"]  # fallback
        elif resp_type == "dict":
            # e.g. {"type":"dict","value":{...}} if your server used that approach
            inner = resp_data.get("value", {})
            return {k: self._deserialize_anything(v) for k, v in inner.items()}

        else:
            raise ValueError(f"Unknown response type: {resp_type}")



import functools
import threading

class CQBase():
    _id_counter = 0
    local_data = threading.local()  # Thread-local storage

    def __init__(self):
        self._instance_id = self._generate_unique_id()

    @classmethod
    def _generate_unique_id(cls):
        with threading.Lock():
            cls._id_counter += 1
            return f"instance_{cls._id_counter}"

    def __getattribute__(self, name):
        attr = super().__getattribute__(name)
        # Intercept callable attributes (methods) and wrap them with decorator
        if callable(attr) and not name.startswith('_'):
            @functools.wraps(attr)
            def intercepted_method(*args, **kwargs):
                #print(f"CLIENT: Intercepted {name} with args={args}, kwargs={kwargs}")
                # Check if the method is being called from a super() call
                if getattr(self.local_data, '_in_super_call', False):
                    # Check if the intercepted method exists on the proxy
                    if hasattr(self.proxy, name):
                        # Call the method remotely using the proxy
                        result = getattr(self.proxy, name)(*args, **kwargs)
                        print(f"SERVER RESULT: {result}")
                        return result
                    else:
                        print(f"CLIENT: {name} not found on server")
                # If not in a super() call, execute the original method locally
                return attr(*args, **kwargs)
            return intercepted_method
        return attr
    
    def __del__(self):
        self.proxy.delete()
        pass


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from gRPC_impl.mytorch.nn import sequential_pb2_grpc, nn_msg_types_pb2
from utils.logger import Logger
from proxies.mytorch.nn.module_proxy import ModuleProxy
from connection_utils.server_connection import wrap_with_error_handler

class SequentialProxy(ModuleProxy):
    def __init__(self, *layers):
        super().__init__()  # Initializes Module, which sets up the module_stub and channel
        self.sequential_stub = sequential_pb2_grpc.SequentialServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.layers = layers
        self.uuid = None

    @wrap_with_error_handler
    def create_sequential_on_server(self):
        layer_descriptions = '\n...'.join([layer.describe() for layer in self.layers])
        self.logger.info(f"Creating Sequential model with layers:\n...{layer_descriptions}")

        # Assuming `layers` is a list of LayerProxy objects or its subclasses (LinearProxy, ReLUProxy, etc.)
        serialized_layers = []

        for layer in self.layers:
            # Create an NNLayer message for each layer
            nn_layer_msg = nn_msg_types_pb2.NNLayer()
            nn_layer_msg.type = layer.layer_type.name
            # Assuming params is a list of strings for simplicity. Adjust based on actual structure.
            # If params need to be key-value pairs serialized as strings, you'll need additional logic to format them.
            for key, value in layer.params.items():
                nn_layer_msg.params.append(f"{key}={value}")

            serialized_layers.append(nn_layer_msg)

        # Create the NNLayers message
        nn_layers_msg = nn_msg_types_pb2.NNLayers()
        nn_layers_msg.layers.extend(serialized_layers)  # Add the NNLayer messages to the NNLayers message

        response = self.sequential_stub.CreateSequentialModuleOnServer(nn_layers_msg)
        self.uuid = response.uuid
        return response.uuid



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class TensorStreamingRequest(_message.Message):
    __slots__ = ("method", "uuid", "two_tensors", "tensor_and_dim", "reshape_request", "slice_request")
    METHOD_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    TWO_TENSORS_FIELD_NUMBER: _ClassVar[int]
    TENSOR_AND_DIM_FIELD_NUMBER: _ClassVar[int]
    RESHAPE_REQUEST_FIELD_NUMBER: _ClassVar[int]
    SLICE_REQUEST_FIELD_NUMBER: _ClassVar[int]
    method: str
    uuid: _shared_msg_types_pb2.UUID
    two_tensors: _shared_msg_types_pb2.TwoTensorIDs
    tensor_and_dim: _shared_msg_types_pb2.TensorIDAndDim
    reshape_request: _shared_msg_types_pb2.ReshapeRequest
    slice_request: SliceRequest
    def __init__(self, method: _Optional[str] = ..., uuid: _Optional[_Union[_shared_msg_types_pb2.UUID, _Mapping]] = ..., two_tensors: _Optional[_Union[_shared_msg_types_pb2.TwoTensorIDs, _Mapping]] = ..., tensor_and_dim: _Optional[_Union[_shared_msg_types_pb2.TensorIDAndDim, _Mapping]] = ..., reshape_request: _Optional[_Union[_shared_msg_types_pb2.ReshapeRequest, _Mapping]] = ..., slice_request: _Optional[_Union[SliceRequest, _Mapping]] = ...) -> None: ...

class TensorStreamingResponse(_message.Message):
    __slots__ = ("tensor", "float_value", "tensor_shape", "tensor_dtype")
    TENSOR_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    TENSOR_SHAPE_FIELD_NUMBER: _ClassVar[int]
    TENSOR_DTYPE_FIELD_NUMBER: _ClassVar[int]
    tensor: _shared_msg_types_pb2.GrpcTensor
    float_value: _shared_msg_types_pb2.FloatValue
    tensor_shape: _shared_msg_types_pb2.TensorShape
    tensor_dtype: _shared_msg_types_pb2.Dtype
    def __init__(self, tensor: _Optional[_Union[_shared_msg_types_pb2.GrpcTensor, _Mapping]] = ..., float_value: _Optional[_Union[_shared_msg_types_pb2.FloatValue, _Mapping]] = ..., tensor_shape: _Optional[_Union[_shared_msg_types_pb2.TensorShape, _Mapping]] = ..., tensor_dtype: _Optional[_Union[_shared_msg_types_pb2.Dtype, _Mapping]] = ...) -> None: ...

class TensorSlice(_message.Message):
    __slots__ = ("start", "stop", "step")
    START_FIELD_NUMBER: _ClassVar[int]
    STOP_FIELD_NUMBER: _ClassVar[int]
    STEP_FIELD_NUMBER: _ClassVar[int]
    start: int
    stop: int
    step: int
    def __init__(self, start: _Optional[int] = ..., stop: _Optional[int] = ..., step: _Optional[int] = ...) -> None: ...

class SliceRequest(_message.Message):
    __slots__ = ("tensor_uuid", "slices")
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    SLICES_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    slices: _containers.RepeatedCompositeFieldContainer[TensorSlice]
    def __init__(self, tensor_uuid: _Optional[str] = ..., slices: _Optional[_Iterable[_Union[TensorSlice, _Mapping]]] = ...) -> None: ...



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GrpcSubset(_message.Message):
    __slots__ = ("uuid", "dataset_length")
    UUID_FIELD_NUMBER: _ClassVar[int]
    DATASET_LENGTH_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    dataset_length: int
    def __init__(self, uuid: _Optional[str] = ..., dataset_length: _Optional[int] = ...) -> None: ...

class GrpcDataLoader(_message.Message):
    __slots__ = ("uuid", "dataset_length", "batch_size")
    UUID_FIELD_NUMBER: _ClassVar[int]
    DATASET_LENGTH_FIELD_NUMBER: _ClassVar[int]
    BATCH_SIZE_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    dataset_length: int
    batch_size: int
    def __init__(self, uuid: _Optional[str] = ..., dataset_length: _Optional[int] = ..., batch_size: _Optional[int] = ...) -> None: ...

class DataLoaderBatchResponse(_message.Message):
    __slots__ = ("inputs_tensor", "targets_tensor", "is_last_batch", "out_of_bounds")
    INPUTS_TENSOR_FIELD_NUMBER: _ClassVar[int]
    TARGETS_TENSOR_FIELD_NUMBER: _ClassVar[int]
    IS_LAST_BATCH_FIELD_NUMBER: _ClassVar[int]
    OUT_OF_BOUNDS_FIELD_NUMBER: _ClassVar[int]
    inputs_tensor: _shared_msg_types_pb2.GrpcTensor
    targets_tensor: _shared_msg_types_pb2.GrpcTensor
    is_last_batch: bool
    out_of_bounds: bool
    def __init__(self, inputs_tensor: _Optional[_Union[_shared_msg_types_pb2.GrpcTensor, _Mapping]] = ..., targets_tensor: _Optional[_Union[_shared_msg_types_pb2.GrpcTensor, _Mapping]] = ..., is_last_batch: bool = ..., out_of_bounds: bool = ...) -> None: ...

class CreateSubsetRequest(_message.Message):
    __slots__ = ("dataset_uuid", "indices")
    DATASET_UUID_FIELD_NUMBER: _ClassVar[int]
    INDICES_FIELD_NUMBER: _ClassVar[int]
    dataset_uuid: str
    indices: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, dataset_uuid: _Optional[str] = ..., indices: _Optional[_Iterable[int]] = ...) -> None: ...

class CreateDataLoaderRequest(_message.Message):
    __slots__ = ("dataset_uuid", "batch_size", "shuffle")
    DATASET_UUID_FIELD_NUMBER: _ClassVar[int]
    BATCH_SIZE_FIELD_NUMBER: _ClassVar[int]
    SHUFFLE_FIELD_NUMBER: _ClassVar[int]
    dataset_uuid: str
    batch_size: int
    shuffle: bool
    def __init__(self, dataset_uuid: _Optional[str] = ..., batch_size: _Optional[int] = ..., shuffle: bool = ...) -> None: ...

class GetNextBatchRequest(_message.Message):
    __slots__ = ("dataloader_uuid", "batch_index")
    DATALOADER_UUID_FIELD_NUMBER: _ClassVar[int]
    BATCH_INDEX_FIELD_NUMBER: _ClassVar[int]
    dataloader_uuid: str
    batch_index: int
    def __init__(self, dataloader_uuid: _Optional[str] = ..., batch_index: _Optional[int] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from .Transformer import Transformer
from proxies.mytorchvision.transforms.ComposeProxy import ComposeProxy

class Compose (Transformer):
    def __init__(self, transforms: list):
        super().__init__()
        self.transforms = transforms
        self.transform_uuids = [transform.uuid for transform in transforms]
        self.proxy = ComposeProxy(self.transform_uuids)
        self.uuid = self.proxy.createTransformOnServer()


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
# Global "singleton" class, for client

import time
from utils.timing_data import TimingData
import uuid
import os
from utils.logger import Logger
import socket
import platform
import requests
import torch
import json

def generate_uuid() -> str:
    return uuid.uuid4().hex # convert UUID to a 32-character string


def get_client_info():
    # Get the hostname
    hostname = socket.gethostname()

    # Get the OS details
    os_name = platform.system()
    os_details = platform.platform()

    # Get the machine type
    machine = platform.machine()
    processor = platform.processor()
    hf_token = os.getenv('HF_TOKEN')
    if hf_token is None:
        # Obsolete usage
        hf_token = os.getenv('HUGGINGFACE_TOKEN')
    if hf_token is None:
        # try to load the token from the .env file in the user's home directory
        hf_token_path = os.path.join(os.path.expanduser('~'), '.cache/huggingface/token')
        if os.path.exists(hf_token_path):
            with open(hf_token_path, 'r') as f:
                hf_token = f.read().strip()
    if hf_token is not None:
        os.environ['HF_TOKEN'] = hf_token
        
    client_info = {
        'hostname': hostname,
        'env': json.dumps(dict(os.environ)),
        'os_name': os_name,
        'os_details': os_details,
        'machine': machine,
        'processor': processor,
        'mytorch_version': torch.__version__,
        'token': os.getenv('MYTORCH_TOKEN'),
    }

    return client_info

class MyTorchClient:
    
    _instance = None
    _current_run = "DEFAULT NAME" 
    _server_ip = 'localhost' # default
    _server_name = 'LOCALHOST' #default
    _server_port = 50051 # default
    _server_addresses = {
        "LOCALHOST": "localhost",
        "TITAN": "10.8.2.17",  # Must be using MyTorch VPN
        }
    _timing_data = TimingData()
    _overwrite_server_data = False
        
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(MyTorchClient, cls).__new__(cls)
            cls._instance._client_id = generate_uuid()
            cls._instance._client_info = get_client_info()
        return cls._instance

    @classmethod
    def instance(cls):
        return cls()
    
    def select_server(self, server_name, server_port=50051) -> tuple[str, str, int] | None: # (server_name, server_ip, server_port)
        server_ip = self._server_addresses[server_name.upper()]
        self._server_ip = server_ip  # self._server_addresses[server_name.upper()]
        self._server_port = server_port
        self._server_name = server_name
        return self._server_name, self._server_ip, self._server_port

    def select_server_by_ip(self, server_ip, server_port=50051):
        # Check if server_ip contains a port specification
        if ':' in server_ip:
            ip_parts = server_ip.split(':')
            server_ip = ip_parts[0]
            server_port = int(ip_parts[1])
        self._server_ip = server_ip  
        self._server_port = server_port
        self._server_name = "N/A"
        return self._server_name, self._server_ip, self._server_port
    
    def select_server_by_proxy(self):
        """Fetches server details from the proxy service."""
        try:
            token = os.getenv('MYTORCH_TOKEN')
            if (token is None or token == ""):
                print("You must add your token to ~/.mytorch or set the MYTORCH_TOKEN environment variable.")
                exit(1)
            response = requests.get(f'https://proxy.mytorch.net/server_request?KEY={token}')

            response.raise_for_status()  # Raises an HTTPError for bad responses
            data = response.json()
            
            self._server_ip = data["SERVER"]
            self._server_port = int(data["PORT"])
            self._server_name = "PROXY"
            if self._server_port == 0: 
                Logger.get_logger().error("All mytorch servers are currently busy. Please try again later.")
                exit(1)
            return self._server_ip, self._server_port
        except Exception as e:
            Logger.get_logger().error(f"Failed to fetch server details from proxy: {str(e)}")
            return None
        
    def select_server_by_env_var(self):
        if not os.getenv('MYTORCH_SERVER_IP'):
            Logger.get_logger().error("MYTORCH_SERVER_IP environment variable not set")
            exit(1)
        server_ip = os.getenv('MYTORCH_SERVER_IP')
        server_port = 50051
        
        # Check if server_ip contains a port specification
        if ':' in server_ip:
            ip_parts = server_ip.split(':')
            server_ip = ip_parts[0]
            server_port = int(ip_parts[1])
        # Use environment port variable if provided and no port in IP
        elif os.getenv('MYTORCH_SERVER_PORT'):
            server_port = int(os.getenv('MYTORCH_SERVER_PORT'))
            
        self._server_ip = server_ip
        self._server_port = server_port
        return self._server_ip, self._server_port

    def get_server(self):
        return self._server_ip, self._server_port

    def get_server_name(self):
        return self._server_name, self._server_port

    def add_method_duration(self, method_name, duration):
        self._timing_data.add_method_duration(self._current_run, method_name, duration)

    def begin_run(self, run_id):
        self._current_run = run_id
        #self._timing_data.clear_statistics()
        start_time = time.perf_counter()
        self._timing_data.start_time(start_time, run_id)

    def end_run(self, run_id, server_stats_json):
        end_time = time.perf_counter()
        self._timing_data.end_time(end_time, run_id)
        server_data = TimingData()  # Temporary instance 
        server_data.init_from_json(server_stats_json, run_id)
        self._timing_data.incorporate_end_of_run_server_data(server_data, self._current_run)

    def get_current_run_id(self):
        return self._current_run
    
    def print_timing_data(self, summary_only=False):
        self._timing_data.print_statistics(summary_only)

    def get_timing_data(self):
        return self._timing_data.get_statistics(self._current_run)

    def clear_timing_data(self, run_id=None):
        self._timing_data.clear_statistics(run_id)

    def set_overwrite_server_data(self, overwrite: bool):
        self._overwrite_server_data = overwrite

    def get_overwrite_server_data(self):
        return self._overwrite_server_data

    def get_client_id(self):
        return self._client_id

    def get_client_info(self):
        return self._client_info



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from proxies.mytorch.nn.loss_function_proxy import LossFunctionProxy
from torch.Tensor import Tensor

class CrossEntropyLoss:
    def __init__(self):
        self.proxy = LossFunctionProxy.create_CrossEntropyLoss()

    # this makes the object 'callable' so that it can be used as a function.
    # For example, we can call `loss_function(input_data, target)` to compute the loss.
    def __call__(self, input_data, target) -> Tensor:
        return self.proxy.run_loss(input_data, target)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: shared_msg_types.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16shared_msg_types.proto\x12\x06shared\"3\n\x16SerializedTensorObject\x12\x19\n\x11serialized_tensor\x18\x02 \x01(\x0c\"B\n\x14SerializedNumpyArray\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\r\n\x05shape\x18\x02 \x03(\x05\x12\r\n\x05\x64type\x18\x03 \x01(\t\"B\n\x14SerializedTensorData\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\r\n\x05shape\x18\x02 \x03(\x05\x12\r\n\x05\x64type\x18\x03 \x01(\t\"\x1d\n\x0cIntegerValue\x12\r\n\x05value\x18\x01 \x01(\x05\"\x1b\n\nFloatValue\x12\r\n\x05value\x18\x01 \x01(\x02\"\x1c\n\x0bStringValue\x12\r\n\x05value\x18\x01 \x01(\t\"(\n\nStringPair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"4\n\x0eReshapeRequest\x12\x13\n\x0btensor_uuid\x18\x01 \x01(\t\x12\r\n\x05shape\x18\x02 \x03(\x05\"\x07\n\x05\x45mpty\"\x1d\n\x0c\x42ooleanValue\x12\r\n\x05value\x18\x01 \x01(\x08\"t\n\x10StringDictionary\x12\x32\n\x05items\x18\x01 \x03(\x0b\x32#.shared.StringDictionary.ItemsEntry\x1a,\n\nItemsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"~\n\x0ckeyValuePair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x16\n\x0cstring_value\x18\x02 \x01(\tH\x00\x12\x13\n\tint_value\x18\x03 \x01(\x05H\x00\x12\x15\n\x0b\x66loat_value\x18\x04 \x01(\x02H\x00\x12\x14\n\nbool_value\x18\x05 \x01(\x08H\x00\x42\x07\n\x05value\"\x14\n\x04UUID\x12\x0c\n\x04uuid\x18\x01 \x01(\t\"\x16\n\x05\x44type\x12\r\n\x05\x64type\x18\x01 \x01(\t\"2\n\x0eTensorIDAndDim\x12\x13\n\x0btensor_uuid\x18\x01 \x01(\t\x12\x0b\n\x03\x64im\x18\x02 \x01(\x05\"4\n\x0fTensorIDsAndDim\x12\x14\n\x0ctensor_uuids\x18\x01 \x03(\t\x12\x0b\n\x03\x64im\x18\x02 \x01(\x05\":\n\x0cTwoTensorIDs\x12\x14\n\x0ctensor1_uuid\x18\x01 \x01(\t\x12\x14\n\x0ctensor2_uuid\x18\x02 \x01(\t\"\x1c\n\x0bTensorShape\x12\r\n\x05shape\x18\x01 \x03(\x05\"8\n\nGrpcTensor\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\r\n\x05shape\x18\x02 \x03(\x05\x12\r\n\x05\x64type\x18\x03 \x01(\t\"?\n\x0bNamedTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\"\n\x06tensor\x18\x02 \x01(\x0b\x32\x12.shared.GrpcTensor\"Z\n\x0eTwoGrpcTensors\x12#\n\x07tensor1\x18\x01 \x01(\x0b\x32\x12.shared.GrpcTensor\x12#\n\x07tensor2\x18\x02 \x01(\x0b\x32\x12.shared.GrpcTensor\":\n\x13MultipleGrpcTensors\x12#\n\x07tensors\x18\x01 \x03(\x0b\x32\x12.shared.GrpcTensor\"i\n\x05Value\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x13\n\tint_value\x18\x02 \x01(\x03H\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x01H\x00\x12\x14\n\nbool_value\x18\x04 \x01(\x08H\x00\x42\x06\n\x04kind\"]\n\tFileChunk\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12\x12\n\nchunk_data\x18\x02 \x01(\x0c\x12\x13\n\x0b\x63hunk_index\x18\x03 \x01(\x05\x12\x15\n\ris_last_chunk\x18\x04 \x01(\x08\"#\n\x0bJsonRequest\x12\x14\n\x0cjson_payload\x18\x01 \x01(\t\"$\n\x0cJsonResponse\x12\x14\n\x0cjson_payload\x18\x01 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'shared_msg_types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STRINGDICTIONARY_ITEMSENTRY']._loaded_options = None
  _globals['_STRINGDICTIONARY_ITEMSENTRY']._serialized_options = b'8\001'
  _globals['_SERIALIZEDTENSOROBJECT']._serialized_start=34
  _globals['_SERIALIZEDTENSOROBJECT']._serialized_end=85
  _globals['_SERIALIZEDNUMPYARRAY']._serialized_start=87
  _globals['_SERIALIZEDNUMPYARRAY']._serialized_end=153
  _globals['_SERIALIZEDTENSORDATA']._serialized_start=155
  _globals['_SERIALIZEDTENSORDATA']._serialized_end=221
  _globals['_INTEGERVALUE']._serialized_start=223
  _globals['_INTEGERVALUE']._serialized_end=252
  _globals['_FLOATVALUE']._serialized_start=254
  _globals['_FLOATVALUE']._serialized_end=281
  _globals['_STRINGVALUE']._serialized_start=283
  _globals['_STRINGVALUE']._serialized_end=311
  _globals['_STRINGPAIR']._serialized_start=313
  _globals['_STRINGPAIR']._serialized_end=353
  _globals['_RESHAPEREQUEST']._serialized_start=355
  _globals['_RESHAPEREQUEST']._serialized_end=407
  _globals['_EMPTY']._serialized_start=409
  _globals['_EMPTY']._serialized_end=416
  _globals['_BOOLEANVALUE']._serialized_start=418
  _globals['_BOOLEANVALUE']._serialized_end=447
  _globals['_STRINGDICTIONARY']._serialized_start=449
  _globals['_STRINGDICTIONARY']._serialized_end=565
  _globals['_STRINGDICTIONARY_ITEMSENTRY']._serialized_start=521
  _globals['_STRINGDICTIONARY_ITEMSENTRY']._serialized_end=565
  _globals['_KEYVALUEPAIR']._serialized_start=567
  _globals['_KEYVALUEPAIR']._serialized_end=693
  _globals['_UUID']._serialized_start=695
  _globals['_UUID']._serialized_end=715
  _globals['_DTYPE']._serialized_start=717
  _globals['_DTYPE']._serialized_end=739
  _globals['_TENSORIDANDDIM']._serialized_start=741
  _globals['_TENSORIDANDDIM']._serialized_end=791
  _globals['_TENSORIDSANDDIM']._serialized_start=793
  _globals['_TENSORIDSANDDIM']._serialized_end=845
  _globals['_TWOTENSORIDS']._serialized_start=847
  _globals['_TWOTENSORIDS']._serialized_end=905
  _globals['_TENSORSHAPE']._serialized_start=907
  _globals['_TENSORSHAPE']._serialized_end=935
  _globals['_GRPCTENSOR']._serialized_start=937
  _globals['_GRPCTENSOR']._serialized_end=993
  _globals['_NAMEDTENSOR']._serialized_start=995
  _globals['_NAMEDTENSOR']._serialized_end=1058
  _globals['_TWOGRPCTENSORS']._serialized_start=1060
  _globals['_TWOGRPCTENSORS']._serialized_end=1150
  _globals['_MULTIPLEGRPCTENSORS']._serialized_start=1152
  _globals['_MULTIPLEGRPCTENSORS']._serialized_end=1210
  _globals['_VALUE']._serialized_start=1212
  _globals['_VALUE']._serialized_end=1317
  _globals['_FILECHUNK']._serialized_start=1319
  _globals['_FILECHUNK']._serialized_end=1412
  _globals['_JSONREQUEST']._serialized_start=1414
  _globals['_JSONREQUEST']._serialized_end=1449
  _globals['_JSONRESPONSE']._serialized_start=1451
  _globals['_JSONRESPONSE']._serialized_end=1487
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytorch.cuda import cuda_pb2 as mytorch_dot_cuda_dot_cuda__pb2
from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytorch/cuda/cuda_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class CudaServiceStub(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.is_available = channel.unary_unary(
                '/mytorch.CudaService/is_available',
                request_serializer=mytorch_dot_cuda_dot_cuda__pb2.is_available_request.SerializeToString,
                response_deserializer=mytorch_dot_cuda_dot_cuda__pb2.is_available_response.FromString,
                _registered_method=True)
        self.empty_cache = channel.unary_unary(
                '/mytorch.CudaService/empty_cache',
                request_serializer=mytorch_dot_cuda_dot_cuda__pb2.empty_cache_request.SerializeToString,
                response_deserializer=mytorch_dot_cuda_dot_cuda__pb2.empty_cache_response.FromString,
                _registered_method=True)
        self.memory_allocated = channel.unary_unary(
                '/mytorch.CudaService/memory_allocated',
                request_serializer=mytorch_dot_cuda_dot_cuda__pb2.memory_allocated_request.SerializeToString,
                response_deserializer=mytorch_dot_cuda_dot_cuda__pb2.memory_allocated_response.FromString,
                _registered_method=True)
        self.memory_reserved = channel.unary_unary(
                '/mytorch.CudaService/memory_reserved',
                request_serializer=mytorch_dot_cuda_dot_cuda__pb2.memory_reserved_request.SerializeToString,
                response_deserializer=mytorch_dot_cuda_dot_cuda__pb2.memory_reserved_response.FromString,
                _registered_method=True)
        self.get_device_properties = channel.unary_unary(
                '/mytorch.CudaService/get_device_properties',
                request_serializer=mytorch_dot_cuda_dot_cuda__pb2.get_device_properties_request.SerializeToString,
                response_deserializer=mytorch_dot_cuda_dot_cuda__pb2.get_device_properties_response.FromString,
                _registered_method=True)
        self.get_device_name = channel.unary_unary(
                '/mytorch.CudaService/get_device_name',
                request_serializer=shared__msg__types__pb2.Empty.SerializeToString,
                response_deserializer=shared__msg__types__pb2.StringValue.FromString,
                _registered_method=True)


class CudaServiceServicer(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def is_available(self, request, context):
        """create a torch.nn.Sequential module on the server and return its UUID to the client
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def empty_cache(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def memory_allocated(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def memory_reserved(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_device_properties(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_device_name(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CudaServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'is_available': grpc.unary_unary_rpc_method_handler(
                    servicer.is_available,
                    request_deserializer=mytorch_dot_cuda_dot_cuda__pb2.is_available_request.FromString,
                    response_serializer=mytorch_dot_cuda_dot_cuda__pb2.is_available_response.SerializeToString,
            ),
            'empty_cache': grpc.unary_unary_rpc_method_handler(
                    servicer.empty_cache,
                    request_deserializer=mytorch_dot_cuda_dot_cuda__pb2.empty_cache_request.FromString,
                    response_serializer=mytorch_dot_cuda_dot_cuda__pb2.empty_cache_response.SerializeToString,
            ),
            'memory_allocated': grpc.unary_unary_rpc_method_handler(
                    servicer.memory_allocated,
                    request_deserializer=mytorch_dot_cuda_dot_cuda__pb2.memory_allocated_request.FromString,
                    response_serializer=mytorch_dot_cuda_dot_cuda__pb2.memory_allocated_response.SerializeToString,
            ),
            'memory_reserved': grpc.unary_unary_rpc_method_handler(
                    servicer.memory_reserved,
                    request_deserializer=mytorch_dot_cuda_dot_cuda__pb2.memory_reserved_request.FromString,
                    response_serializer=mytorch_dot_cuda_dot_cuda__pb2.memory_reserved_response.SerializeToString,
            ),
            'get_device_properties': grpc.unary_unary_rpc_method_handler(
                    servicer.get_device_properties,
                    request_deserializer=mytorch_dot_cuda_dot_cuda__pb2.get_device_properties_request.FromString,
                    response_serializer=mytorch_dot_cuda_dot_cuda__pb2.get_device_properties_response.SerializeToString,
            ),
            'get_device_name': grpc.unary_unary_rpc_method_handler(
                    servicer.get_device_name,
                    request_deserializer=shared__msg__types__pb2.Empty.FromString,
                    response_serializer=shared__msg__types__pb2.StringValue.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytorch.CudaService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytorch.CudaService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class CudaService(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    @staticmethod
    def is_available(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.CudaService/is_available',
            mytorch_dot_cuda_dot_cuda__pb2.is_available_request.SerializeToString,
            mytorch_dot_cuda_dot_cuda__pb2.is_available_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def empty_cache(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.CudaService/empty_cache',
            mytorch_dot_cuda_dot_cuda__pb2.empty_cache_request.SerializeToString,
            mytorch_dot_cuda_dot_cuda__pb2.empty_cache_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def memory_allocated(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.CudaService/memory_allocated',
            mytorch_dot_cuda_dot_cuda__pb2.memory_allocated_request.SerializeToString,
            mytorch_dot_cuda_dot_cuda__pb2.memory_allocated_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def memory_reserved(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.CudaService/memory_reserved',
            mytorch_dot_cuda_dot_cuda__pb2.memory_reserved_request.SerializeToString,
            mytorch_dot_cuda_dot_cuda__pb2.memory_reserved_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def get_device_properties(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.CudaService/get_device_properties',
            mytorch_dot_cuda_dot_cuda__pb2.get_device_properties_request.SerializeToString,
            mytorch_dot_cuda_dot_cuda__pb2.get_device_properties_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def get_device_name(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.CudaService/get_device_name',
            shared__msg__types__pb2.Empty.SerializeToString,
            shared__msg__types__pb2.StringValue.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class LossFunctionType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    MSE: _ClassVar[LossFunctionType]
    L1: _ClassVar[LossFunctionType]
    CrossEntropy: _ClassVar[LossFunctionType]
    NLL: _ClassVar[LossFunctionType]
    PoissonNLL: _ClassVar[LossFunctionType]
    KLDiv: _ClassVar[LossFunctionType]
    BCE: _ClassVar[LossFunctionType]
    MarginRanking: _ClassVar[LossFunctionType]
    Hinge: _ClassVar[LossFunctionType]
    MultiLabelMargin: _ClassVar[LossFunctionType]
    SmoothL1: _ClassVar[LossFunctionType]
    SoftMargin: _ClassVar[LossFunctionType]
    MultiLabelSoftMargin: _ClassVar[LossFunctionType]
    CosineEmbedding: _ClassVar[LossFunctionType]
    MultiMargin: _ClassVar[LossFunctionType]
    TripletMargin: _ClassVar[LossFunctionType]
MSE: LossFunctionType
L1: LossFunctionType
CrossEntropy: LossFunctionType
NLL: LossFunctionType
PoissonNLL: LossFunctionType
KLDiv: LossFunctionType
BCE: LossFunctionType
MarginRanking: LossFunctionType
Hinge: LossFunctionType
MultiLabelMargin: LossFunctionType
SmoothL1: LossFunctionType
SoftMargin: LossFunctionType
MultiLabelSoftMargin: LossFunctionType
CosineEmbedding: LossFunctionType
MultiMargin: LossFunctionType
TripletMargin: LossFunctionType

class LossFunctionRequest(_message.Message):
    __slots__ = ("loss_function_type",)
    LOSS_FUNCTION_TYPE_FIELD_NUMBER: _ClassVar[int]
    loss_function_type: LossFunctionType
    def __init__(self, loss_function_type: _Optional[_Union[LossFunctionType, str]] = ...) -> None: ...

class RunLossFunctionRequest(_message.Message):
    __slots__ = ("loss_function_uuid", "input_uuid", "target_uuid")
    LOSS_FUNCTION_UUID_FIELD_NUMBER: _ClassVar[int]
    INPUT_UUID_FIELD_NUMBER: _ClassVar[int]
    TARGET_UUID_FIELD_NUMBER: _ClassVar[int]
    loss_function_uuid: str
    input_uuid: str
    target_uuid: str
    def __init__(self, loss_function_uuid: _Optional[str] = ..., input_uuid: _Optional[str] = ..., target_uuid: _Optional[str] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

# Import for direct access
from .mytorch import *

from . import nn
from . import cuda
from . import optim
from . import hub
from . import scaffolding
from . import jit
from . import scaffolding
from .scaffolding import select_server_by_ip, select_server

# Make sure everything is accessible when mytorch is imported as torch
__all__ = ['randn', 'nn', 'from_numpy', 'cuda', 'no_grad', 'optim', 'max', 'Tensor', 'rand',
           'arange', 'meshgrid', 'reshape', 'cat', 'argmax', 'matmul', 'load', 'save', 'hub', 'scaffolding']


# __version__ = "0.1.0" # Initial version; dotplot example
# __version__ = "0.1.5" # ImageFolder infer demo
#__version__ = "0.2.0" # Added Resnet training demo
#__version__ = "0.2.1" # Added uploading data to server during inference (Zynq edge demo)
#__version__ = "0.2.2" # Tracing demo (jit.trace)
#__version__ = "0.2.3" # llama3 CLI demo (Hugging Face transformers library)
#__version__ = "0.2.4" # Pass up client ID to server for data storage
#__version__ = "0.2.5" # Fixed mem leak; --gpu_stats flag
#__version__ = "0.2.6" # Server: multi-client; Client: centralized error handling
#__version__ = "0.2.7" # Basic AWS Neuron support
#__version__ = "0.2.8" # Reorg of grpc_impl
#__version__ = "0.3.0" # Rename to mytorch
#__version__ = "0.3.1" # Added functionality to login to huggingface on the server side
#__version__ = "0.4.0" # Major change due to footbag's generic call supporting anomaly detection
#__version__ = "0.4.1" # Had to increase the required version number of grpc
__version__ = "0.4.2"  # Forgot to modify setup_mytorch.py with new grpc version number



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorchvision.datasets import datasets_pb2, datasets_pb2_grpc

class CIFAR10Proxy:
    def __init__(self, root: str, train: bool, download: bool, transform_uuid: str):
        self.channel = ServerConnection.get_active_connection()
        self.root = root
        self.train = train
        self.download = download
        self.transform_uuid = transform_uuid
        self.stub = datasets_pb2_grpc.DatasetsServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.uuid = None

    @wrap_with_error_handler
    def createCIFAR10OnServer(self):
        request = datasets_pb2.CreateCIFAR10Request(root=self.root, train=self.train, download=self.download, transform_uuid=self.transform_uuid)
        response: datasets_pb2.GrpcCIFAR10 = self.stub.CreateCIFAR10onServer(request)
        self.uuid = response.uuid
        return response.uuid, response.dataset_length, response.classes


from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from gRPC_impl.mytorch.nn import nn_msg_types_pb2 as _nn_msg_types_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ModuleToDeviceRequest(_message.Message):
    __slots__ = ("uuid", "device")
    UUID_FIELD_NUMBER: _ClassVar[int]
    DEVICE_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    device: str
    def __init__(self, uuid: _Optional[str] = ..., device: _Optional[str] = ...) -> None: ...

class Parameters(_message.Message):
    __slots__ = ("generator_uuid",)
    GENERATOR_UUID_FIELD_NUMBER: _ClassVar[int]
    generator_uuid: str
    def __init__(self, generator_uuid: _Optional[str] = ...) -> None: ...

class half_request(_message.Message):
    __slots__ = ("uuid",)
    UUID_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    def __init__(self, uuid: _Optional[str] = ...) -> None: ...

class half_response(_message.Message):
    __slots__ = ("empty",)
    EMPTY_FIELD_NUMBER: _ClassVar[int]
    empty: _shared_msg_types_pb2.Empty
    def __init__(self, empty: _Optional[_Union[_shared_msg_types_pb2.Empty, _Mapping]] = ...) -> None: ...



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class is_available_request(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class is_available_response(_message.Message):
    __slots__ = ("is_available",)
    IS_AVAILABLE_FIELD_NUMBER: _ClassVar[int]
    is_available: bool
    def __init__(self, is_available: bool = ...) -> None: ...

class empty_cache_request(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class empty_cache_response(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class memory_allocated_request(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class memory_allocated_response(_message.Message):
    __slots__ = ("memory_allocated",)
    MEMORY_ALLOCATED_FIELD_NUMBER: _ClassVar[int]
    memory_allocated: int
    def __init__(self, memory_allocated: _Optional[int] = ...) -> None: ...

class memory_reserved_request(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class memory_reserved_response(_message.Message):
    __slots__ = ("memory_reserved",)
    MEMORY_RESERVED_FIELD_NUMBER: _ClassVar[int]
    memory_reserved: int
    def __init__(self, memory_reserved: _Optional[int] = ...) -> None: ...

class get_device_properties_request(_message.Message):
    __slots__ = ("device_str", "device_int")
    DEVICE_STR_FIELD_NUMBER: _ClassVar[int]
    DEVICE_INT_FIELD_NUMBER: _ClassVar[int]
    device_str: str
    device_int: int
    def __init__(self, device_str: _Optional[str] = ..., device_int: _Optional[int] = ...) -> None: ...

class get_device_properties_response(_message.Message):
    __slots__ = ("name", "total_memory", "available_memory", "allocated_memory", "reserved_memory", "max_memory_allocated", "max_memory_reserved", "multi_processor_count")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TOTAL_MEMORY_FIELD_NUMBER: _ClassVar[int]
    AVAILABLE_MEMORY_FIELD_NUMBER: _ClassVar[int]
    ALLOCATED_MEMORY_FIELD_NUMBER: _ClassVar[int]
    RESERVED_MEMORY_FIELD_NUMBER: _ClassVar[int]
    MAX_MEMORY_ALLOCATED_FIELD_NUMBER: _ClassVar[int]
    MAX_MEMORY_RESERVED_FIELD_NUMBER: _ClassVar[int]
    MULTI_PROCESSOR_COUNT_FIELD_NUMBER: _ClassVar[int]
    name: str
    total_memory: int
    available_memory: int
    allocated_memory: int
    reserved_memory: int
    max_memory_allocated: int
    max_memory_reserved: int
    multi_processor_count: int
    def __init__(self, name: _Optional[str] = ..., total_memory: _Optional[int] = ..., available_memory: _Optional[int] = ..., allocated_memory: _Optional[int] = ..., reserved_memory: _Optional[int] = ..., max_memory_allocated: _Optional[int] = ..., max_memory_reserved: _Optional[int] = ..., multi_processor_count: _Optional[int] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/scaffolding/data_mgmt.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#mytorch/scaffolding/data_mgmt.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"6\n\x11\x46ileUploadRequest\x12\x10\n\x08\x66ilepath\x18\x01 \x01(\t\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\x0c\"%\n\x14\x46olderUploadResponse\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\"\x1c\n\x0cRelativePath\x12\x0c\n\x04path\x18\x01 \x01(\t\"\x17\n\x07\x44\x61taDir\x12\x0c\n\x04path\x18\x01 \x01(\t2\xd8\x01\n\x0f\x44\x61taMgmtService\x12K\n\x0cUploadFolder\x12\x1a.mytorch.FileUploadRequest\x1a\x1d.mytorch.FolderUploadResponse(\x01\x12\x43\n\x14\x46olderExistsOnServer\x12\x15.mytorch.RelativePath\x1a\x14.shared.BooleanValue\x12\x33\n\x10SetServerDataDir\x12\x10.mytorch.DataDir\x1a\r.shared.Emptyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.scaffolding.data_mgmt_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FILEUPLOADREQUEST']._serialized_start=72
  _globals['_FILEUPLOADREQUEST']._serialized_end=126
  _globals['_FOLDERUPLOADRESPONSE']._serialized_start=128
  _globals['_FOLDERUPLOADRESPONSE']._serialized_end=165
  _globals['_RELATIVEPATH']._serialized_start=167
  _globals['_RELATIVEPATH']._serialized_end=195
  _globals['_DATADIR']._serialized_start=197
  _globals['_DATADIR']._serialized_end=220
  _globals['_DATAMGMTSERVICE']._serialized_start=223
  _globals['_DATAMGMTSERVICE']._serialized_end=439
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from torchvision.transforms.Transformer import Transformer
from proxies.mytorchvision.datasets import CIFAR10Proxy
from torch.utils.data.Dataset import Dataset

class CIFAR10(Dataset):
    def __init__(self, root: str, train: bool, download: bool, transform: Transformer):
        super().__init__()
        self.root = root
        self.train = train
        self.download = download
        self.transform = transform
        self.proxy = CIFAR10Proxy(root, train, download, self.transform.uuid)
        self.uuid, self.dataset_length, self.classes = self.createCIFAR10onServer()

    def createCIFAR10onServer(self):
        uuid, dataset_length, classes = self.proxy.createCIFAR10OnServer()
        return uuid, dataset_length, list(classes)

    def __len__(self):
        return self.dataset_length



###############################################################################
# Copyright (c) 2025 MyTorch Systems Inc. All rights reserved.
###############################################################################
#from torch.nn.Module import Module
#from proxies.mytorch.mytorch_proxy import MyTorchProxy
#from torch.Tensor import Tensor
from proxies.mytorch.nn.layers_proxy import SigmoidProxy

class Sigmoid:
    
    def __init__(self):
        self.proxy = SigmoidProxy()

    def __call__(self, tensor):
        # Delegate the call to SigmoidProxy for execution
        return self.proxy.forward(tensor)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/nn/loss_functions.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fmytorch/nn/loss_functions.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"L\n\x13LossFunctionRequest\x12\x35\n\x12loss_function_type\x18\x01 \x01(\x0e\x32\x19.mytorch.LossFunctionType\"]\n\x16RunLossFunctionRequest\x12\x1a\n\x12loss_function_uuid\x18\x01 \x01(\t\x12\x12\n\ninput_uuid\x18\x02 \x01(\t\x12\x13\n\x0btarget_uuid\x18\x03 \x01(\t*\x87\x02\n\x10LossFunctionType\x12\x07\n\x03MSE\x10\x00\x12\x06\n\x02L1\x10\x01\x12\x10\n\x0c\x43rossEntropy\x10\x02\x12\x07\n\x03NLL\x10\x03\x12\x0e\n\nPoissonNLL\x10\x04\x12\t\n\x05KLDiv\x10\x05\x12\x07\n\x03\x42\x43\x45\x10\x06\x12\x11\n\rMarginRanking\x10\x07\x12\t\n\x05Hinge\x10\x08\x12\x14\n\x10MultiLabelMargin\x10\t\x12\x0c\n\x08SmoothL1\x10\n\x12\x0e\n\nSoftMargin\x10\x0b\x12\x18\n\x14MultiLabelSoftMargin\x10\x0c\x12\x13\n\x0f\x43osineEmbedding\x10\r\x12\x0f\n\x0bMultiMargin\x10\x0e\x12\x11\n\rTripletMargin\x10\x0f\x32\x9f\x01\n\x13LossFunctionService\x12@\n\x12\x43reateLossFunction\x12\x1c.mytorch.LossFunctionRequest\x1a\x0c.shared.UUID\x12\x46\n\x0fRunLossFunction\x12\x1f.mytorch.RunLossFunctionRequest\x1a\x12.shared.GrpcTensorb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.nn.loss_functions_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_LOSSFUNCTIONTYPE']._serialized_start=242
  _globals['_LOSSFUNCTIONTYPE']._serialized_end=505
  _globals['_LOSSFUNCTIONREQUEST']._serialized_start=68
  _globals['_LOSSFUNCTIONREQUEST']._serialized_end=144
  _globals['_RUNLOSSFUNCTIONREQUEST']._serialized_start=146
  _globals['_RUNLOSSFUNCTIONREQUEST']._serialized_end=239
  _globals['_LOSSFUNCTIONSERVICE']._serialized_start=508
  _globals['_LOSSFUNCTIONSERVICE']._serialized_end=667
# @@protoc_insertion_point(module_scope)



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class ResourceRequest(_message.Message):
    __slots__ = ("auth_token", "client_ip", "server_id", "max_time")
    AUTH_TOKEN_FIELD_NUMBER: _ClassVar[int]
    CLIENT_IP_FIELD_NUMBER: _ClassVar[int]
    SERVER_ID_FIELD_NUMBER: _ClassVar[int]
    MAX_TIME_FIELD_NUMBER: _ClassVar[int]
    auth_token: str
    client_ip: str
    server_id: str
    max_time: int
    def __init__(self, auth_token: _Optional[str] = ..., client_ip: _Optional[str] = ..., server_id: _Optional[str] = ..., max_time: _Optional[int] = ...) -> None: ...

class ResourceRequestResponse(_message.Message):
    __slots__ = ("access_token", "additional_info", "resource_id", "resource_ip", "resource_port", "resource_name", "lease_time", "mytorch_version", "resource_info")
    ACCESS_TOKEN_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_INFO_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_IP_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_PORT_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_NAME_FIELD_NUMBER: _ClassVar[int]
    LEASE_TIME_FIELD_NUMBER: _ClassVar[int]
    MYTORCH_VERSION_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_INFO_FIELD_NUMBER: _ClassVar[int]
    access_token: str
    additional_info: str
    resource_id: str
    resource_ip: str
    resource_port: int
    resource_name: str
    lease_time: int
    mytorch_version: str
    resource_info: str
    def __init__(self, access_token: _Optional[str] = ..., additional_info: _Optional[str] = ..., resource_id: _Optional[str] = ..., resource_ip: _Optional[str] = ..., resource_port: _Optional[int] = ..., resource_name: _Optional[str] = ..., lease_time: _Optional[int] = ..., mytorch_version: _Optional[str] = ..., resource_info: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/nn/nn_msg_types.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dmytorch/nn/nn_msg_types.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"\x9f\x01\n\x13SerializedStateDict\x12:\n\x07\x65ntries\x18\x01 \x03(\x0b\x32).mytorch.SerializedStateDict.EntriesEntry\x1aL\n\x0c\x45ntriesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.shared.SerializedTensorData:\x02\x38\x01\"\'\n\x07NNLayer\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0e\n\x06params\x18\x02 \x03(\t\",\n\x08NNLayers\x12 \n\x06layers\x18\x01 \x03(\x0b\x32\x10.mytorch.NNLayer\">\n\x12\x46orwardPassRequest\x12\x13\n\x0bmodule_uuid\x18\x01 \x01(\t\x12\x13\n\x0btensor_uuid\x18\x02 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.nn.nn_msg_types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SERIALIZEDSTATEDICT_ENTRIESENTRY']._loaded_options = None
  _globals['_SERIALIZEDSTATEDICT_ENTRIESENTRY']._serialized_options = b'8\001'
  _globals['_SERIALIZEDSTATEDICT']._serialized_start=67
  _globals['_SERIALIZEDSTATEDICT']._serialized_end=226
  _globals['_SERIALIZEDSTATEDICT_ENTRIESENTRY']._serialized_start=150
  _globals['_SERIALIZEDSTATEDICT_ENTRIESENTRY']._serialized_end=226
  _globals['_NNLAYER']._serialized_start=228
  _globals['_NNLAYER']._serialized_end=267
  _globals['_NNLAYERS']._serialized_start=269
  _globals['_NNLAYERS']._serialized_end=313
  _globals['_FORWARDPASSREQUEST']._serialized_start=315
  _globals['_FORWARDPASSREQUEST']._serialized_end=377
# @@protoc_insertion_point(module_scope)



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class KeyValuePair(_message.Message):
    __slots__ = ("key", "value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class ClientInfo(_message.Message):
    __slots__ = ("key_value_pairs",)
    KEY_VALUE_PAIRS_FIELD_NUMBER: _ClassVar[int]
    key_value_pairs: _containers.RepeatedCompositeFieldContainer[KeyValuePair]
    def __init__(self, key_value_pairs: _Optional[_Iterable[_Union[KeyValuePair, _Mapping]]] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024-2025 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.optim.adam_proxy import AdamProxy
from typing import Iterable, Dict, Any, Tuple, Union, Optional
from torch.Tensor import Tensor
from typing_extensions import TypeAlias

ParamsT: TypeAlias = Union[
    Iterable[Tensor], Iterable[Dict[str, Any]], Iterable[Tuple[str, Tensor]]
]

class Adam:
    def __init__(
        self,
        params: ParamsT,
        lr: Union[float, Tensor] = 1e-3,
        betas: Tuple[Union[float, Tensor], Union[float, Tensor]] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        amsgrad: bool = False,
        *,
        foreach: Optional[bool] = None,
        maximize: bool = False,
        capturable: bool = False,
        differentiable: bool = False,
        fused: Optional[bool] = None,
        ):
            self.proxy = AdamProxy()

            param_uuids = []
            for p in params:
                if hasattr(p, 'uuid'):
                    param_uuids.append(p.uuid)
                else:
                    raise TypeError(f"Adam received a non-proxied parameter: {p}")

            self.proxy.uuid = self.proxy.generic_call(
                "torch.optim", "Adam",
                param_uuids,  
                call_type="constructor",
                kwargs={
                    "lr": lr,
                    "betas": betas,
                    "eps": eps,
                    "weight_decay": weight_decay,
                    "amsgrad": amsgrad,
                    "foreach": foreach,
                    "maximize": maximize,
                    "capturable": capturable,
                    "differentiable": differentiable,
                    "fused": fused
                }
            )


    def zero_grad(self):
        return self.proxy.zero_grad()

    def step(self):
        return self.proxy.step()


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/nn/functional/functional.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&mytorch/nn/functional/functional.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"2\n\x0eSoftmaxRequest\x12\x13\n\x0btensor_uuid\x18\x01 \x01(\t\x12\x0b\n\x03\x64im\x18\x02 \x01(\x05\x32K\n\x11\x46unctionalService\x12\x36\n\x07softmax\x12\x17.mytorch.SoftmaxRequest\x1a\x12.shared.GrpcTensorb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.nn.functional.functional_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SOFTMAXREQUEST']._serialized_start=75
  _globals['_SOFTMAXREQUEST']._serialized_end=125
  _globals['_FUNCTIONALSERVICE']._serialized_start=127
  _globals['_FUNCTIONALSERVICE']._serialized_end=202
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This proxy provides a proxy for the MyTorchService gRPC service.
It allows the client to call specified torch operations on the server via gRPC.
"""

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch.cuda import cuda_pb2_grpc, cuda_pb2
from gRPC_impl.shared_msg_types_pb2 import Empty
from utils.logger import Logger
from proxies.base_proxy import BaseProxy


class CudaProxy (BaseProxy):

    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = cuda_pb2_grpc.CudaServiceStub(self.channel)
        self.logger = Logger.get_logger()

    #@wrap_with_error_handler
    #def is_available(self) -> bool:
    #    response: cuda_pb2.is_available_response = self.stub.is_available(cuda_pb2.is_available_request())
    #    return response.is_available
    
    @wrap_with_error_handler
    def is_available(self) -> bool:
        is_avail = self.generic_call("torch.cuda", "is_available")
        return is_avail

    #@wrap_with_error_handler
    #def get_device_name(self) -> str:
    #    response: cuda_pb2.get_device_name_response = self.stub.get_device_name(Empty())
    #    return response.value

    @wrap_with_error_handler
    def get_device_name(self) -> str:
        response= self.generic_call("torch.cuda", "get_device_name")
        return response

    @wrap_with_error_handler
    def empty_cache(self):
        self.stub.empty_cache(cuda_pb2.empty_cache_request())

    @wrap_with_error_handler
    def memory_allocated(self) -> int:
        response: cuda_pb2.memory_allocated_response = self.stub.memory_allocated(cuda_pb2.memory_allocated_request())
        return response.memory_allocated

    @wrap_with_error_handler
    def memory_reserved(self) -> int:
        response: cuda_pb2.memory_reserved_response = self.stub.memory_reserved(cuda_pb2.memory_reserved_request())
        return response.memory_reserved

    @wrap_with_error_handler
    def get_device_properties(self, device) -> dict:
        if isinstance(device, str):
            request = cuda_pb2.get_device_properties_request(device_str=device)
        elif isinstance(device, int):
            request = cuda_pb2.get_device_properties_request(device_int=device)
        else:
            self.logger.error("No device specified")
            return {}
        response: cuda_pb2.get_device_properties_response = self.stub.get_device_properties(request)
        return {
            "name": response.name,
            "total_memory": response.total_memory,
            "available_memory": response.available_memory,
            "allocated_memory": response.allocated_memory,
            "reserved_memory": response.reserved_memory,
            "max_memory_allocated": response.max_memory_allocated,
            "max_memory_reserved": response.max_memory_reserved,
            "multi_processor_count": response.multi_processor_count
        }



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SerializedStateDict(_message.Message):
    __slots__ = ("entries",)
    class EntriesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _shared_msg_types_pb2.SerializedTensorData
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_shared_msg_types_pb2.SerializedTensorData, _Mapping]] = ...) -> None: ...
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    entries: _containers.MessageMap[str, _shared_msg_types_pb2.SerializedTensorData]
    def __init__(self, entries: _Optional[_Mapping[str, _shared_msg_types_pb2.SerializedTensorData]] = ...) -> None: ...

class NNLayer(_message.Message):
    __slots__ = ("type", "params")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    type: str
    params: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, type: _Optional[str] = ..., params: _Optional[_Iterable[str]] = ...) -> None: ...

class NNLayers(_message.Message):
    __slots__ = ("layers",)
    LAYERS_FIELD_NUMBER: _ClassVar[int]
    layers: _containers.RepeatedCompositeFieldContainer[NNLayer]
    def __init__(self, layers: _Optional[_Iterable[_Union[NNLayer, _Mapping]]] = ...) -> None: ...

class ForwardPassRequest(_message.Message):
    __slots__ = ("module_uuid", "tensor_uuid")
    MODULE_UUID_FIELD_NUMBER: _ClassVar[int]
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    module_uuid: str
    tensor_uuid: str
    def __init__(self, module_uuid: _Optional[str] = ..., tensor_uuid: _Optional[str] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

import logging


# Map string representations to logging level constants
level_map = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARNING,
    'ERROR': logging.ERROR
}

MIN_LOG_LEVEL = logging.INFO

class Logger:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Logger, cls).__new__(cls)
            logger = logging.getLogger()
            logger.setLevel(MIN_LOG_LEVEL)
            cls._instance.logger = logger

            # Define formats
            debug_format = logging.Formatter('%(message)s')
            general_format = logging.Formatter('%(levelname)s - %(message)s')

            # Stream handler for debug level
            debug_handler = logging.StreamHandler()
            debug_handler.setLevel(logging.DEBUG)
            debug_handler.setFormatter(debug_format)
            debug_handler.addFilter(logging.Filter())  # Allow only DEBUG
            debug_handler.addFilter(lambda record: record.levelno == logging.DEBUG)

            # Stream handler for all levels, but configured to handle INFO and above
            general_handler = logging.StreamHandler()
            general_handler.setLevel(logging.INFO)  # Changed to INFO
            general_handler.setFormatter(general_format)

            # Add handlers to the logger
            logger.addHandler(debug_handler)
            logger.addHandler(general_handler)

        return cls._instance

    @staticmethod
    def get_logger():
        return Logger()._instance.logger



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/nn/sequential.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2
from gRPC_impl.mytorch.nn import nn_msg_types_pb2 as mytorch_dot_nn_dot_nn__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bmytorch/nn/sequential.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\x1a\x1dmytorch/nn/nn_msg_types.proto2V\n\x11SequentialService\x12\x41\n\x1e\x43reateSequentialModuleOnServer\x12\x11.mytorch.NNLayers\x1a\x0c.shared.UUIDb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.nn.sequential_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SEQUENTIALSERVICE']._serialized_start=95
  _globals['_SEQUENTIALSERVICE']._serialized_end=181
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch.hub import hub_pb2, hub_pb2_grpc
from torchvision.models import WeightsEnum
from utils.logger import Logger

class HubProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = hub_pb2_grpc.HubServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.uuid = None

    # example call: torch.hub.load('pytorch/vision:v0.6.0', 'resnet18', weights=ResNet18_Weights.DEFAULT)
    @wrap_with_error_handler
    def load_module_from_repo(self, repo: str, model_name: str, weights: WeightsEnum) -> str:
        # Load a model from the PyTorch Hub
        request = hub_pb2.ModelLoadRequest(repo=repo, model_name=model_name, weights_enum=weights.name)
        response = self.stub.loadModel(request)
        self.logger.info(f"Loaded model from repo {repo} with model name {model_name}")
        return response.model_uuid